# -*- coding: utf-8 -*-
from ..errors import XmppError
from ..stanzas import Stanza


NS_URI = "urn:ietf:params:xml:ns:xmpp-tls"


async def handle(stream, feature_elem, timeout=None):
    nsmap = {"tls": NS_URI}

    stream.send(Stanza("starttls", nsmap={None: NS_URI}))
    resp = await stream.wait([("/tls:proceed", nsmap),
                              ("/tls:failure", nsmap)], timeout=timeout)
    if resp.name == "{%s}proceed" % NS_URI:
        await stream._transport.starttls()
    else:
        # A real stanza/stream error type is not wrapped by the <failure/>,
        # unlike other newer protocols, so gin up a dummy.
        raise XmppError("starttls failure: %s" % resp.toXml().decode())

    return True


def isRequired(feature_elem):
    return ("{%s}required" % NS_URI) in [c.tag for c in feature_elem]



# -*- coding: utf-8 -*-
import time
from lxml import etree

from ..stanzas import Iq
from ..stream import Mixin
from ..utils import xpathFilter

from .. import getLogger
log = getLogger(__name__)

"""
Protocol for XEP 0202 - Entity Time support.
http://www.xmpp.org/extensions/xep-0202.html
"""

NS_URI = "urn:xmpp:time"
GET_XPATH = ("/iq[@type='get']/ns:time", {"ns": NS_URI})


async def get(stream, to):
    iq = await stream.sendAndWaitIq(NS_URI, to=to, child_name="time",
                                    raise_on_error=True)
    return iq


class EntityTimeMixin(Mixin):

    @xpathFilter(GET_XPATH)
    async def onStanza(self, stream, stanza):
        log.debug("%s received" % NS_URI)

        result = Iq(xml=stanza.xml)
        result.swapToFrom()
        result.type = "result"

        elem = etree.Element("tzo")
        timezone = time.timezone if not time.localtime().tm_isdst \
                                 else time.altzone
        timezone = (-1 * timezone) if time.localtime().tm_gmtoff < 0 \
                                   else timezone
        elem.text = "%+03d:%02d" % (timezone / 60, abs(timezone % 60))
        result.request.append(elem)

        elem = etree.Element("utc")
        elem.text = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
        result.request.append(elem)

        stream.send(result)



# -*- coding: utf-8 -*-
from nicfit.aio import Application
from nicfit.console import pout
from . import version


async def main(args):
    pout("\m/")


app = Application(main, version=version,
                  gettext_domain=None)

if __name__ == "__main__":
    app.run()



# -*- coding: utf-8 -*-
from nose.tools import *
from vexmpp.stanzas import *

def testPresenceOrdering():
    assert_less(Presence(priority=5), Presence(priority=10))
    assert_less(Presence(priority=0), Presence(priority=1))
    assert_less(Presence(priority=-1), Presence(priority=0))
    assert_less(Presence(priority=-127), Presence(priority=126))

    assert_less(Presence(priority=5, show=Presence.SHOW_CHAT),
                Presence(priority=10))

    assert_greater(Presence(priority=10, show=Presence.SHOW_CHAT),
                   Presence(priority=10, show=Presence.SHOW_AWAY))
    assert_greater(Presence(priority=10, show=Presence.SHOW_CHAT),
                   Presence(priority=10, show=None))
    assert_greater(Presence(priority=10, show=None),
                   Presence(priority=10, show=Presence.SHOW_AWAY))
    assert_greater(Presence(priority=10, show=Presence.SHOW_AWAY),
                   Presence(priority=10, show=Presence.SHOW_XA))
    assert_greater(Presence(priority=10, show=Presence.SHOW_XA),
                   Presence(priority=10, show=Presence.SHOW_DND))

    assert_false(Presence(priority=5) < Presence(priority=5))
    assert_false(Presence(priority=5) > Presence(priority=5))
    assert_false(Presence(priority=10, show=Presence.SHOW_CHAT) <
                 Presence(priority=10, show=Presence.SHOW_CHAT))
    assert_false(Presence(priority=10, show=Presence.SHOW_CHAT) >
                 Presence(priority=10, show=Presence.SHOW_CHAT))



# -*- coding: utf-8 -*-

"""
test_jid
----------------------------------

Tests for `vexmpp.xmpp.jid` module.
"""
import unittest
from vexmpp.jid import Jid, _parse, _prep, InvalidJidError, internJid


class TestJidParsing(unittest.TestCase):

    def setUp(self):
        pass

    def test_basic(self):
        self.assertEqual(_parse("user@host/resource"),
                          ("user", "host", "resource"))
        self.assertEqual(_parse("user@host"),
                          ("user", "host", None))
        self.assertEqual(_parse("host"),
                          (None, "host", None))
        self.assertEqual(_parse("host/resource"),
                          (None, "host", "resource"))

        self.assertEqual(_parse("foo/bar@baz"),
                          (None, "foo", "bar@baz"))
        self.assertEqual(_parse("boo@foo/bar@baz"),
                          ("boo", "foo", "bar@baz"))
        self.assertEqual(_parse("boo@foo/bar/baz"),
                          ("boo", "foo", "bar/baz"))
        self.assertEqual(_parse("boo/foo@bar@baz"),
                          (None, "boo", "foo@bar@baz"))
        self.assertEqual(_parse("boo/foo/bar"),
                          (None, "boo", "foo/bar"))
        self.assertEqual(_parse("boo//foo"),
                          (None, "boo", "/foo"))

    def test_noHost(self):
        '''
        Test for failure on no host part.
        '''
        self.assertRaises(InvalidJidError, _parse, "user@")

    def test_doubleAt(self):
        """
        Test for failure on double @ signs.

        This should fail because @ is not a valid character for the host
        part of the JID.
        """
        self.assertRaises(UnicodeError, _parse, "user@@host")

    def test_multipleAt(self):
        """
        Test for failure on two @ signs.

        This should fail because @ is not a valid character for the host
        part of the JID.
        """
        self.assertRaises(UnicodeError, _parse, "user@host@host")

    # Basic tests for case mapping. These are fallback tests for the
    # prepping done in twisted.words.protocols.jabber.xmpp_stringprep

    def test_prepCaseMapUser(self):
        """
        Test case mapping of the user part of the JID.
        """
        self.assertEqual(_prep("UsEr", "host", "resource"),
                          ("user", "host", "resource"))

    def test_prepCaseMapHost(self):
        """
        Test case mapping of the host part of the JID.
        """
        self.assertEqual(_prep("user", "hoST", "resource"),
                          ("user", "host", "resource"))

    def test_prepNoCaseMapResource(self):
        """
        Test no case mapping of the resourcce part of the JID.
        """
        self.assertEqual(_prep("user", "hoST", "resource"),
                          ("user", "host", "resource"))
        self.assertNotEqual(_prep("user", "host", "Resource"),
                            ("user", "host", "resource"))

    def tearDown(self):
        pass


class TestJidObject(unittest.TestCase):

    def setUp(self):
        pass

    def test_ctor_types(self):
        self.assertRaises(ValueError, Jid, b"bytes")

        self.assertRaises(ValueError, Jid, (b"user", "host", "rsrc"))
        self.assertRaises(ValueError, Jid, ("user", b"host", "rsrc"))
        self.assertRaises(ValueError, Jid, ("user", "host", b"rsrc"))

    def tearDown(self):
        pass

    def test_noneArguments(self):
        """
        Test that using no arguments raises an exception.
        """
        self.assertRaises(TypeError, Jid)

    def test_attributes(self):
        """
        Test that the attributes correspond with the JID parts.
        """
        j = Jid("user@host/resource")
        self.assertEqual(j.user, "user")
        self.assertEqual(j.host, "host")
        self.assertEqual(j.resource, "resource")

    def test_userhost(self):
        """
        Test the extraction of the bare JID.
        """
        j = Jid("user@host/resource")
        self.assertEqual("user@host", j.bare)

    def test_userhostOnlyHost(self):
        """
        Test the extraction of the bare JID of the full form host/resource.
        """
        j = Jid("host/resource")
        self.assertEqual("host", j.bare)

    def test_userhostJID(self):
        """
        Test getting a JID object of the bare JID.
        """
        j1 = Jid("user@host/resource")
        j2 = internJid("user@host")
        self.assertEqual(id(j2), id(j1.bare_jid))

    def test_userhostJIDNoResource(self):
        """
        Test getting a JID object of the bare JID when there was no resource.
        """
        j = Jid("user@host")
        self.assertEqual(id(j), id(j.bare_jid))

    def test_fullHost(self):
        """
        Test giving a string representation of the JID with only a host part.
        """
        j = Jid((None, 'host', None))
        self.assertEqual('host', j.full)

    def test_fullHostResource(self):
        """
        Test giving a string representation of the JID with host, resource.
        """
        j = Jid((None, 'host', 'resource'))
        self.assertEqual('host/resource', j.full)

    def test_fullUserHost(self):
        """
        Test giving a string representation of the JID with user, host.
        """
        j = Jid(('user', 'host', None))
        self.assertEqual('user@host', j.full)

    def test_fullAll(self):
        """
        Test giving a string representation of the JID.
        """
        j = Jid(('user', 'host', 'resource'))
        self.assertEqual('user@host/resource', j.full)

    def test_equality(self):
        """
        Test JID equality.
        """
        j1 = Jid("user@host/resource")
        j2 = Jid("user@host/resource")
        self.assertNotEqual(id(j1), id(j2))
        self.assertEqual(j1, j2)

    def test_equalityWithNonJIDs(self):
        """
        Test JID equality.
        """
        j = Jid("user@host/resource")
        try:
            res = (j == "user@host/resource")
        except NotImplementedError:
            pass
        else:
            self.assertFalse("Jid and strings should not be comparable")

    def test_inequality(self):
        """
        Test JID inequality.
        """
        j1 = Jid("user1@host/resource")
        j2 = Jid("user2@host/resource")
        self.assertNotEqual(j1, j2)

    def test_inequalityWithNonJIDs(self):
        """
        Test JID equality.
        """
        j = Jid("user@host/resource")
        self.assertNotEqual(j, 'user@host/resource')

    def test_hashable(self):
        """
        Test JID hashability.
        """
        j1 = Jid("user@host/resource")
        j2 = Jid("user@host/resource")
        self.assertEqual(hash(j1), hash(j2))

    def test_unicode(self):
        """
        Test unicode representation of JIDs.
        """
        j = Jid(('user', 'host', 'resource'))
        self.assertEqual("user@host/resource", j.full)

    def test_repr(self):
        """
        Test representation of JID objects.
        """
        j = Jid(('user', 'host', 'resource'))
        self.assertEqual("Jid('user@host/resource')", repr(j))



# -*- coding: utf-8 -*-
from ..protocols import resourcebind, sessionbind


async def handle(stream, feature_elem, timeout=None):
    jid = await resourcebind.bind(stream, stream.creds.jid.resource,
                                  timeout=timeout)
    if jid.resource != stream.creds.jid.resource:
        # Server generated the resource (or changed it!)
        stream.creds.jid = jid

    # The 'session' feature is in 3921 but NOT 6120
    if feature_elem.getparent().xpath("child::sess:session",
                                      namespaces={"sess":
                                                  sessionbind.NS_URI}):
        # FIXME: only do this if session is in the features
        # Start the session
        await sessionbind.newsession(stream, timeout=timeout)



# -*- coding: utf-8 -*-
from ..stream import Mixin
from ..utils import xpathFilter

from .. import getLogger
log = getLogger(__name__)

NS_URI = "urn:xmpp:sm:3"


class Opts():
    resume = False
    sm_id = None
    resume_location = None
    max_resume_time = None


class Mixin(Mixin):
    '''XXX: this is a very rudimentary XEP-0198 Mixin. Currently all it does
    is count stanza in and out.'''

    def __init__(self, opts=None):
        self._opts = opts or Opts()
        self._in_count = 0
        self._out_count = 0
        super().__init__([('stream_mgmt_opts', self._opts)])

    @xpathFilter(["/iq", "/presence", "/message",
                  ("/ns:a", {"ns": NS_URI}), ("/ns:r", {"ns": NS_URI}),
                 ])
    async def onStanza(self, stream, stanza):
        self._in_count += 1
        log.debug("Stanza IN count: {:d}".format(self._in_count))

    @xpathFilter(["/iq", "/presence", "/message"])
    def onSend(self, stream, stanza):
        self._out_count += 1
        log.debug("Stanza OUT count: {:d}".format(self._out_count))



# -*- coding: utf-8 -*-
import os
import asyncio
from hashlib import md5
from pathlib import Path
from tempfile import mkstemp
from base64 import b64decode, b64encode
from collections import namedtuple

from lxml import etree

from .. import getLogger
from ..errors import (BadRequestStanzaError, XmppError,
                      PolicyViolationStanzaError,
                      FeatureNotImplementedStanzaError)
from ..protocols import xdata
from ..stanzas import Iq

log = getLogger(__name__)

SI_NS_URI = "http://jabber.org/protocol/si"
FT_NS_URI = "http://jabber.org/protocol/si/profile/file-transfer"
FEATURE_NS_URI = "http://jabber.org/protocol/feature-neg"

IBB_NS_URI = "http://jabber.org/protocol/ibb"
BYTESTREAMS_NS_URI = "http://jabber.org/protocol/bytestreams"

RangeTuple = namedtuple("RangeTuple", "offset, length")
# FIXME: Collapse into File class. Lazy open temp file
FileInfo = namedtuple("FileInfo", "name, size, desc, date, md5, range, sid")

MAX_SEQ = 65535
BLOCK_SIZE = 4096


class FileTransferError(Exception):
    pass


class File:
    def __init__(self, info, maxsize=None, tempdir=None):
        self.info = info
        self._data_size = 0
        self._next_seq = None
        self._seq_set = set()
        self._md5sum = md5()
        self._maxsize = maxsize

        self._tmp_dir = Path(tempdir).resolve()
        (self._tmp_fd,
         self._tmp_name) = mkstemp(suffix=".incoming", dir=str(self._tmp_dir))
        log.debug("Opened tmp file {}".format(self._tmp_name))

    def addBlock(self, data):
        if not self._tmp_fd:
            raise ValueError("File object has not been opened.")

        if self._next_seq is None:
            self._next_seq = 0

        try:
            seq = int(data.get("seq"))
            if seq in self._seq_set:
                raise ValueError("Duplicate data sequence.")
            elif seq != self._next_seq:
                raise ValueError("Data sequence out of order.")

            self._seq_set.add(seq)
            self._next_seq += 1

            bytes_ = b64decode(data.text)
            num_bytes = len(bytes_)
            self._data_size += num_bytes
            if self._maxsize and self._data_size > self._maxsize:
                raise OverflowError("File exceeds size limit of {:d} bytes."
                                    .format(self._maxsize))
        except:
            self.close(False)
            raise

        os.write(self._tmp_fd, bytes_)
        self._md5sum.update(bytes_)
        log.info("Writing {} bytes for seq {}".format(num_bytes, seq))

        # TODO: handle MAX_SEQ rollover

    def _cleanupTemp(self):
        if os.path.isfile(self._tmp_name):
            os.remove(self._tmp_name)
        self._tmp_name = None

    def _verify(self):
        sequences = sorted(self._seq_set)
        if len(sequences) != sequences[-1] + 1:
            # Missing sequences
            full_set = set(range(sequences[-1] + 1))
            if full_set.difference(self._seq_set):
                raise FileTransferError(
                    "Not all data blocks received, missing sequences {}"
                    .format(full_set.difference(self._seq_set)))

        if self._data_size != self.info.size:
            raise FileTransferError("Not all bytes received, read {:d} bytes "
                                    "but expected {:d}".format(self._data_size,
                                                               self.info.size))

        if self.info.md5:
            md5 = self.md5
            if self.info.md5.lower() != md5:
                raise FileTransferError("File hash mismatch, computed {} but "
                                        "expected {}".format(md5,
                                                             self.info.md5))

    def close(self, success):
        close_err = None

        if self._tmp_fd:
            os.close(self._tmp_fd)
            log.debug("Closed tmp file {}".format(self._tmp_name))
            self._tmp_fd = None

        if success:
            try:
                self._verify()
            except Exception as ex:
                close_err = ex

        if not success or close_err:
            self._cleanupTemp()

        if close_err:
            raise close_err

    @property
    def md5(self):
        return self._md5sum.hexdigest()

    @property
    def filename(self):
        return self._tmp_name


async def receiveFile(stream, si_request, maxsize=None, tempdir=None,
                      timeout=None):
    no_valid_streams_error = etree.Element("no-valid-streams",
                                           nsmap={None: SI_NS_URI})

    si_elem = si_request.find("./si:si", namespaces={"si": SI_NS_URI})
    try:
        if si_elem.attrib["profile"] != FT_NS_URI:
            raise BadRequestStanzaError("Profile not understood",
                    app_err=etree.Element("bad-profile",
                                          nsmap={None: SI_NS_URI}))
        sid = si_elem.attrib["id"]

        file_elem = si_request.find(".//f:file", namespaces={"f": FT_NS_URI})
        form = si_request.find(".//f:feature/x:x[@type='form']",
                            namespaces={"x": "jabber:x:data",
                                        "f": FEATURE_NS_URI})
        if file_elem is None:
            raise BadRequestStanzaError("no <file> child")
        elif form is None:
            raise BadRequestStanzaError("no x-data form")

        form = xdata.XdataForm(form)
        field = form.field("stream-method")
        if field is None:
            raise BadRequestStanzaError("no stream-method field",
                                        app_err=no_valid_streams_error)

        # XXX: Field class with nice accessors,  etc..
        methods = field.xpath("./x:option/x:value/text()",
                              namespaces={"x": xdata.NS_URI})

        if (IBB_NS_URI not in methods):
            raise BadRequestStanzaError("Only in-band-bytestreams is supported",
                                    app_err=no_valid_streams_error)
    except XmppError as err:
        stream.send(si_request.errorResponse(err))
        raise FileTransferError(str(err))

    # TODO: check and prefer bytesstream here.
    selected_meth = IBB_NS_URI
    receiveFileCoro = ibbReceiveFile
    open_xpath = ("/iq[@type='set']/ibb:open[@sid='{}']".format(sid),
                  {"ibb": IBB_NS_URI})

    range_elem = file_elem.find("./{%s}range" % FT_NS_URI)
    range_info = None
    if range_elem is not None:
        r_offset = range_elem.get("offset")
        r_length = range_elem.get("length")
        if (r_offset, r_length) != (None, None):
            range_info = RangeTuple(offset=int(r_offset), length=int(r_length))

    file_info = FileInfo(name=file_elem.get("name"),
                         size=int(file_elem.get("size")),
                         desc=file_elem._getChildText("desc"),
                         date=file_elem.get("date"),
                         md5=file_elem.get("hash"),
                         range=range_info,
                         sid=sid)

    form.type = "submit"
    field.clear()
    field.set("var", "stream-method")
    e = etree.Element("value")
    e.text = selected_meth
    field.append(e)

    response = si_request.resultResponse()
    stream.send(response)

    open_iq = await stream.wait(open_xpath, timeout=timeout)
    new_file = await receiveFileCoro(stream, open_iq, file_info,
                                     maxsize=maxsize, tempdir=tempdir,
                                     timeout=timeout)
    return new_file


async def sendFile(stream, to_jid, filename, description=None, timeout=None):
    request = Iq(to=to_jid, type="set", request=("si", SI_NS_URI))
    sid = request.id

    h = md5()
    path = Path(filename)
    h.update(path.open("rb").read())

    # FIXME: date .. need a xmpp date format util
    file_info = FileInfo(name=path.name, size=path.stat().st_size,
                         desc=description, date=None, md5=h.hexdigest(),
                         range=None, sid=sid)

    si = request.query
    si.attrib["id"] = sid
    si.attrib["profile"] = FT_NS_URI

    f = si.appendChild("file", FT_NS_URI)
    f.attrib["name"] = file_info.name
    f.attrib["size"] = str(file_info.size)
    if description:
        f.appendChild("desc").text = description
    if file_info.md5:
        f.attrib["hash"] = file_info.md5
    if file_info.date:
        f.attrib["date"] = file_info.date
    # TODO: ranges

    feat = si.appendChild("feature", FEATURE_NS_URI)
    form = xdata.XdataForm()
    form.appendListField("stream-method", [IBB_NS_URI], default=IBB_NS_URI)
    feat.append(form)

    # Initiate stream
    try:
        response = await stream.sendAndWait(request, timeout=timeout,
                                            raise_on_error=True)
    except XmppError as ex:
        raise FileTransferError(ex)

    si_elem = response.find("./si:si", namespaces={"si": SI_NS_URI})
    methods = si_elem.xpath(".//x:field/x:value/text()",
                            namespaces={"x": xdata.NS_URI})
    if not methods or (methods[0] not in [IBB_NS_URI]):
        stream.send(response.errrorResponse(BadRequestStanzaError(
                "Only in-band-bytestreams is supported",
                app_err=etree.Element("no-valid-streams",
                                      nsmap={None: SI_NS_URI}))))
        raise FileTransferError("No supported stream methods.")

    # TODO: check and prefer bytesstream here.
    selected_meth = IBB_NS_URI
    sendFileCoro = ibbSendFile

    open_iq = Iq(to=to_jid, type="set", request=("open", selected_meth))
    open_iq.query.attrib["sid"] = sid
    open_iq.query.attrib["block-size"] = str(BLOCK_SIZE)
    open_iq.query.attrib["stanza"] = "iq"

    try:
        await stream.sendAndWait(open_iq, raise_on_error=True,
                                 timeout=timeout)
        await sendFileCoro(stream, to_jid, path, sid, timeout=timeout)
    except XmppError as ex:
        raise FileTransferError(ex)


# IBB ---------------------------------------------------------------------

async def ibbSendFile(stream, to_jid, file_path, sid, block_size=BLOCK_SIZE,
                      timeout=None):
    assert(isinstance(file_path, Path))

    open_iq = Iq(to=to_jid, type="set", request=("open", IBB_NS_URI))
    open_iq.query.attrib["sid"] = sid
    open_iq.query.attrib["block-size"] = str(block_size)
    open_iq.query.attrib["stanza"] = "iq"

    await stream.sendAndWait(open_iq, raise_on_error=True, timeout=timeout)

    with file_path.open("rb") as fp:
        seq = 0
        while True:
            data = fp.read(BLOCK_SIZE)
            if not data:
                break
            iq = Iq(to=to_jid, type="set", request=("data", IBB_NS_URI))
            iq.query.attrib["sid"] = sid
            iq.query.attrib["seq"] = str(seq)
            iq.query.text = b64encode(data)
            await stream.sendAndWait(iq, raise_on_error=True, timeout=timeout)
            seq += 1

    close_iq = Iq(to=to_jid, type="set", request=("close", IBB_NS_URI))
    close_iq.query.attrib["sid"] = sid
    await stream.sendAndWait(close_iq, raise_on_error=True, timeout=timeout)


async def ibbReceiveFile(stream, request, file_info, maxsize=None,
                         tempdir=None, timeout=None):

    open_elem = request.getChild("open", IBB_NS_URI)
    if open_elem is None:
        raise BadRequestStanzaError("no <open>")

    nsmap = {"ibb": IBB_NS_URI}

    from_jid = request.frm
    block_size = open_elem.get("block-size")
    sid = open_elem.get("sid")
    stanza = open_elem.get("iq")
    if block_size is None or sid is None:
        raise BadRequestStanzaError("Missing block-size and/or sid")
    elif stanza == "message":
        raise FeatureNotImplementedStanzaError("IBB via messages not supported")

    done = False
    response = request.resultResponse()
    stream.send(response)

    ft_file = File(file_info, maxsize, tempdir=tempdir)

    # Read data
    while not done:
        data_xp = "/iq[@from='{frm}' and @type='set']"\
                  "/ibb:data[@sid='{sid}']"\
                  .format(frm=from_jid.full, sid=sid)
        close_xp = "/iq[@from='{frm}' and @type='set']"\
                   "/ibb:close[@sid='{sid}']"\
                   .format(frm=from_jid.full, sid=sid)

        try:
            iq = await stream.wait([(data_xp, nsmap), (close_xp, nsmap)],
                                   timeout=timeout)
        except asyncio.TimeoutError:
            ft_file.close(False)
            raise

        close = iq.getChild("close", IBB_NS_URI)
        if close is not None:
            ft_file.close(True)
            done = True
        else:
            data = iq.getChild("data", IBB_NS_URI)
            try:
                ft_file.addBlock(data)
            except (ValueError, OverflowError) as ex:
                m = str(ex)
                if isinstance(ex, OverflowError):
                    xmpp_error = PolicyViolationStanzaError(m)
                else:
                    xmpp_error = BadRequestStanzaError(m)
                stream.send(iq.errorResponse(xmpp_error))
                raise FileTransferError(m)

        response = iq.resultResponse(clear=True)
        stream.send(response)

    return ft_file



from ..util import bytes
from ..sasl import Mechanism, register_mechanism


class X_FACEBOOK_PLATFORM(Mechanism):

    def __init__(self, sasl, name):
        super(X_FACEBOOK_PLATFORM, self).__init__(sasl, name)
        self.check_values(['access_token', 'api_key'])

    def process(self, challenge=None):
        if challenge is not None:
            values = {}
            for kv in challenge.split(b'&'):
                key, value = kv.split(b'=')
                values[key] = value

            resp_data = {
                b'method': values[b'method'],
                b'v': b'1.0',
                b'call_id': b'1.0',
                b'nonce': values[b'nonce'],
                b'access_token': self.values['access_token'],
                b'api_key': self.values['api_key']
            }
            resp = '&'.join(['%s=%s' % (k, v) for k, v in resp_data.items()])
            return bytes(resp)
        return b''

    def okay(self):
        return True


register_mechanism('X-FACEBOOK-PLATFORM', 40, X_FACEBOOK_PLATFORM,
                   use_hashes=False)



# -*- coding: utf-8 -*-
from threading import Lock

#
# Portions of this code stolen from Twisted, their licence applies there.
# _parse, _prep
# Copyright (c) 2001-2005 Twisted Matrix Laboratories.
# See LICENSE for details.
#

__internJIDs = {}
__internJIDs_lock = Lock()


class InvalidJidError(RuntimeError):
    '''Exception type for Invalid Jabber Identifiers.'''


class Jid(object):
    def __init__(self, j):
        if isinstance(j, str):
            (self._local_part,
             self._domain_part,
             self._rsrc_part) = _parse(j)
        elif isinstance(j, tuple):
            (self._local_part,
             self._domain_part,
             self._rsrc_part) = _prep(*j)
        elif issubclass(j.__class__, Jid):
            (self._local_part,
             self._domain_part,
             self._rsrc_part) = (j._local_part, j._domain_part, j._rsrc_part)
        else:
            raise ValueError("Type must be str or "
                             "tuple(str, str, str). "
                             "Got '%s' instead." % str(type(j)))

    @property
    def user(self):
        return self._local_part
    localpart = user
    '''Alias for user'''

    @property
    def host(self):
        return self._domain_part
    domainpart = host
    '''Alias for host'''

    @property
    def resource(self):
        return self._rsrc_part
    resourcepart = resource
    '''Alias for resource'''

    @property
    def full(self):
        if self._local_part:
            if self._rsrc_part:
                return "%s@%s/%s" % (self._local_part,
                                      self._domain_part,
                                      self._rsrc_part)
            else:
                return "%s@%s" % (self._local_part, self._domain_part)
        else:
            if self.resource:
                return "%s/%s" % (self._domain_part, self._rsrc_part)
            else:
                return self._domain_part

    @property
    def bare(self):
        if self._local_part:
            return "%s@%s" % (self._local_part, self._domain_part)
        else:
            return self._domain_part

    def __eq__(self, other):
        if isinstance(other, Jid):
            return (self.user == other.user and
                    self.host == other.host and
                    self.resource == other.resource)
        else:
            raise NotImplementedError("Jid and '%s' not compariable" %
                                      str(type(other)))

    @property
    def bare_jid(self):
        """
        Extract the bare JID.

        A bare JID does not have a resource part, so this returns a
        Jid object representing either user@host or just host.

        If the JID object doesn't have a resource set, it will return itself
        (i.e. self). Otherwise, the bare JID object will be created, interned
        using ``internJID``.
        """
        if self.resource:
            return internJid(self.bare)
        else:
            return self

    def __ne__(self, other):
        try:
            result = self.__eq__(other)
        except NotImplementedError:
            return True
        return not result

    def __hash__(self):
        return hash((self.user, self.host, self.resource))

    def __unicode__(self):
        return self.full

    def __repr__(self):
        return 'Jid(%r)' % self.full


def _parse(j):
    user = None
    host = None
    resource = None

    # Search for delimiters
    user_sep = j.find("@")
    res_sep = j.find("/")

    if user_sep == -1:
        if res_sep == -1:
            # host
            host = j
        else:
            # host/resource
            host = j[0:res_sep]
            resource = j[res_sep + 1:] or None
    else:
        if res_sep == -1:
            # user@host
            user = j[0:user_sep] or None
            host = j[user_sep + 1:]
        else:
            if user_sep < res_sep:
                # user@host/resource
                user = j[0:user_sep] or None
                host = j[user_sep + 1:user_sep + (res_sep - user_sep)]
                resource = j[res_sep + 1:] or None
            else:
                # host/resource (with an @ in resource)
                host = j[0:res_sep]
                resource = j[res_sep + 1:] or None

    return _prep(user, host, resource)


def _prep(user, host, resource):
    from ._stringprep import nodeprep, nameprep, resourceprep

    for arg in (user, host, resource):
        if arg and not isinstance(arg, str):
            raise ValueError("Jid comonents must be unicode strings")

    if user:
        user = nodeprep.prepare(user)
    else:
        user = None

    if not host:
        raise InvalidJidError("Domain part required.")
    else:
        host = nameprep.prepare(host)

    if resource:
        resource = resourceprep.prepare(resource)
    else:
        resource = None

    return (user, host, resource)


def internJid(jidstring):
    """
    Return interned JID.
    """

    __internJIDs_lock.acquire()
    try:
        if jidstring in __internJIDs:
            return __internJIDs[jidstring]
        else:
            j = Jid(jidstring)
            # Recheck the map, the normalized form may actually be there.
            jidstring = j.full
            if jidstring in __internJIDs:
                return __internJIDs[jidstring]
            else:
                __internJIDs[jidstring] = j
                return j
    finally:
        __internJIDs_lock.release()



from .anonymous import ANONYMOUS  # noqa
from .plain import PLAIN  # noqa
from .cram_md5 import CRAM_MD5  # noqa
from .digest_md5 import DIGEST_MD5  # noqa
from .scram_hmac import SCRAM_HMAC  # noqa
from .messenger_oauth2 import X_MESSENGER_OAUTH2  # noqa
from .facebook_platform import X_FACEBOOK_PLATFORM  # noqa



# -*- coding: utf-8 -*-
import math
from collections import OrderedDict

from . import getLogger
log = getLogger(__name__)


class BaseMetric:
    def __init__(self, name, type_=int):
        self.name = name
        self._value = type_()
        self._type = type_

    @property
    def value(self):
        return self._value

    def update(self, val):
        raise NotImplementedError()

    def __str__(self):
        return "<{name}>{_value}</{name}>".format(**self.__dict__)


class CounterMetric(BaseMetric):
    def update(self, val):
        self._value = val

    def inc(self, val=1):
        self._value += val

    def dec(self, val=1):
        self._value -= val


class ValueMetric(BaseMetric):
    def __init__(self, name, initial_val=None, type_=None):
        if type_ is None:
            if initial_val is not None:
                type_ = type(type_)
            else:
                type_ = int
        super().__init__(name, type_)

        self._count = 0
        self._min, self._max = None, None
        self._total = 0
        self._total_squared = 0
        if initial_val is not None:
            self.update(initial_val)

    def update(self, val):
        self._total += val
        self._total_squared += val * val
        self._min = min(val, self._min) if self._min is not None else val
        self._max = max(val, self._max) if self._max is not None else val

        self._count += 1
        self._value = val

    @property
    def min(self):
        return self._min

    @property
    def max(self):
        return self._max

    @property
    def average(self):
        try:
            return self._total / self._count
        except ZeroDivisionError:
            return None

    # http://stackoverflow.com/questions/5543651/
    #      computing-standard-deviation-in-a-stream
    @property
    def stdev(self):
        s0 = self._count
        s1 = self._total
        s2 = self._total_squared
        try:
            return math.sqrt((s0 * s2 - s1 * s1) / (s0 * (s0 - 1)))
        except ZeroDivisionError:
            return None


class MetricsGroup(object):
    def __init__(self, name):
        self.name = name
        self._metrics = OrderedDict()

    def add(self, s):
        self._metrics[s.name] = s
        return s

    @property
    def metrics(self):
        return [m for m in self._metrics]

    def __str__(self):
        stats = ''.join([str(s) for s in self._metrics.values()])
        return "<{name}>{stats}</{name}>".format(name=self.name, stats=stats)

    def __getitem__(self, key):
        return self._metrics[key]

    def valueDict(self):
        metric_vals = {}
        for n, v in self._metrics.items():
            metric_vals[n] = v.value
        return metric_vals

    @classmethod
    def fromString(clazz, s):
        instance = clazz()

        # currently XML, meh
        from lxml import etree
        try:
            metrics_elem = etree.fromstring(s)
            metrics_elem = metrics_elem.xpath("/metrics/%s" % clazz.__name__)[0]
            for elem in metrics_elem.getchildren():
                m = instance[elem.tag]
                m.update(m._type(elem.text))
        except Exception as ex:
            raise ValueError("{} : {}".format(s, str(ex)))

        return instance



# -*- coding: utf-8 -*-
from ..stanzas import Stanza
from ..errors import makeStanzaError
from ..protocols.stream_mgmt import NS_URI


async def handle(stream, feature_elem, sm_opts, timeout=None):
    assert(feature_elem is not None)
    nsmap = {"sm": NS_URI}

    enable_elem = Stanza("enable", nsmap={None: NS_URI})
    if sm_opts and sm_opts.resume:
        enable_elem.set("resume", "true")
    stream.send(enable_elem)

    resp = await stream.wait([("/sm:enabled", nsmap),
                              ("/sm:failed", nsmap)], timeout=timeout)
    if resp.name == "{%s}failed" % NS_URI:
        raise makeStanzaError(resp.xml)

    sm_opts.sm_id = resp.get("id")
    sm_opts.resume = bool(resp.get("resume") and
                          resp.get("resume") in ("1", "true"))
    sm_opts.resume_location = resp.get("location")
    sm_opts.max_resume_time = resp.get("max")

    return True



# -*- coding: utf-8 -*-
from ..stanzas import Iq

'''
Vcard

http://xmpp.org/extensions/xep-0054.html
http://xmpp.org/extensions/xep-0153.html
'''

NS_URI = "vcard-temp"


async def get(stream, to, timeout=None):
    iq = await stream.sendAndWaitIq(NS_URI, to=to, child_name="vCard",
                                    raise_on_error=True, id_prefix="vcard_get",
                                    timeout=timeout)
    # TODO: ensure vCard child
    return iq


async def set(stream, to, vcard_xml, timeout=None):
    iq = Iq(to=to, type="set", id_prefix="vcard_set")
    iq.append(vcard_xml)
    iq = await stream.sendAndWait(iq, raise_on_error=True, timeout=timeout)
    return iq

# TODO: vCard builder



# -*- coding: utf-8 -*-
import base64
from lxml.etree import Element
from nicfit import getLogger

from .. import suelta
from ..protocols.sasl import SaslError, NS_URI


async def handle(stream, feature_elem, timeout=None):
    nsmap = {"sasl": NS_URI}

    server_mechs = []
    for mech_elem in feature_elem.xpath("//sasl:mechanism",
                                        namespaces=nsmap):
        server_mechs.append(mech_elem.text)

    def _tls_active():
        return stream.tls_active

    def _callback(mech, vals):
        log.debug("SASL _callback mech: %s values: %s" % (mech, vals))
        if "password" in vals:
            vals["password"] = stream.creds.password
        mech.fulfill(vals)

    def _secquery(mech, question):
        # FIXME: what is this?
        # import ipdb; ipdb.set_trace()
        return True

    # This could be set to a desired mech. From a config, for example.
    MECH = None
    sasl_engine = suelta.SASL(stream.creds.jid.host, "xmpp",
                              username=stream.creds.jid.user,
                              mech=MECH,
                              sec_query=_secquery,
                              tls_active=_tls_active,
                              request_values=_callback)
    sasl_mech = sasl_engine.choose_mechanism(server_mechs)

    auth = Element("{%s}auth" % NS_URI, nsmap={None: NS_URI})
    auth.attrib["mechanism"] = sasl_mech.name
    initial_resp = sasl_mech.process(None)
    if initial_resp:
        log.debug("initial_resp: %s" % str(initial_resp))
        auth.text = base64.b64encode(initial_resp)

    stream.send(auth)

    done = False
    while not done:
        resp = await stream.wait([("/sasl:success", nsmap),
                                  ("/sasl:failure", nsmap),
                                  ("/sasl:challenge", nsmap),
                                 ], timeout=timeout)
        if resp.xml.tag == "{%s}success" % NS_URI:
            done = True
            if resp.xml.text:
                # XXX: not really sure what to do with this
                log.debug("<success/> txt: %s" %
                          base64.b64decode(resp.xml.text))
        elif resp.xml.tag == "{%s}failure" % NS_URI:
            raise SaslError(xml=resp.xml)
        else:
            assert(resp.xml.tag == "{%s}challenge" % NS_URI)
            challenge = base64.b64decode(resp.xml.text.strip())
            log.debug("challenge: %s" % challenge)
            resp_txt = sasl_mech.process(challenge)
            auth = Element("{%s}response" % NS_URI, nsmap={None: NS_URI})
            if resp_txt:
                auth.text = base64.b64encode(resp_txt)
            stream.send(auth)


log = getLogger(__name__)



import os
import asyncio
import logging

from functools import partial
from collections import deque

from lxml import etree

from . import stanzas
from .stanzas import Iq
from .parser import Parser
from .utils import signalEvent
from .utils import benchmark as timedWait

from . import getLogger
log = getLogger(__name__)


if "VEX_TIMED_WAITS" in os.environ and int(os.environ["VEX_TIMED_WAITS"]):
    from .metrics import ValueMetric
    stream_wait_met = ValueMetric("stream:wait_time", type_=float)
else:
    stream_wait_met = None

_ENFORCE_TIMEOUTS = bool("VEX_ENFORCE_TIMEOUTS" in os.environ and
                         int(os.environ["VEX_ENFORCE_TIMEOUTS"]))


class QueuedStanza:
    def __init__(self, s):
        self.task_set = set()
        self.stanza = s


class ParserTask(asyncio.Task):
    def __init__(self, stream, loop=None):
        super().__init__(self._run(), loop=loop)
        self._parser = Parser()
        self._data_queue = asyncio.Queue()
        self._stream = stream

    def parse(self, bytes_):
        self._data_queue.put_nowait(bytes_)

    def reset(self):
        self._parser.reset()

    async def _run(self):
        while True:
            try:
                data = await self._data_queue.get()

                elems = self._parser.parse(data)
                for e in elems:
                    stanza = stanzas.makeStanza(e)
                    if log.getEffectiveLevel() <= logging.VERBOSE:
                        log.verbose("[STANZA IN]:\n%s" %
                                    stanza.toXml(pprint=True).decode("utf-8"))
                    await self._stream._handleStanza(stanza)
            except Exception as ex:
                log.exception(ex)


class Stream(asyncio.Protocol):
    """Base class for XMPP streams."""

    def __init__(self, creds, state_callbacks=None, mixins=None,
                 default_timeout=None):
        self.creds = creds
        self._transport = None
        self._waiter_futures = []
        self._tls_active = False
        self._callbacks = state_callbacks

        self._mixins = mixins or []
        for mixin in self._mixins:
            for name, obj in mixin._exports:
                if name in self.__dict__:
                    raise ValueError("Mixin '%s' exports ambiguous "
                                     "data named '%s'" % (str(mixin), name))
                else:
                    # Add the symbol to the stream's namespace
                    self.__dict__[name] = obj

        self._parser_task = ParserTask(self)
        self.default_timeout = default_timeout
        # Stream errors
        self.error = None

        self._stanza_queue = deque(maxlen=10)

    @property
    def connected(self):
        if not self._transport:
            return False
        else:
            if (getattr(self._transport, "_closing") and
                    self._transport._closing):
                # SSL transport
                return False

        return True

    @property
    def tls_active(self):
        return self._tls_active

    @property
    def jid(self):
        return self.creds.jid

    def close(self):
        if self.connected:
            self.send(b"</stream:stream>")
            self._transport.close()

    def send(self, data):
        """Send ``data`` which can be a vexmpp.stanza.Stanza,
         lxml.etree.Element, a str, or bytes. The the case of bytes the
         encoding MUST be utf-8 encoded (per XMPP specification).

         In the case of Stanza and Element the Mixin.onSend callback is
         invoked. Currently there is not a Mixin callback for strings or bytes.
        """
        def _send(bytes_):
            if not self._transport:
                log.warn("Data send with disconnected transport")
                return

            self._transport.write(bytes_)
            log.debug("[BYTES OUT]: %s", bytes_)

        stanza = None
        if isinstance(data, stanzas.Stanza):
            stanza = data
            raw_data = data.toXml()
        elif isinstance(data, str):
            raw_data = data.encode("utf-8")
        elif isinstance(data, etree._Element):
            stanza = stanzas.Stanza(xml=data)
            raw_data = etree.tostring(data, encoding="utf-8")
        elif isinstance(data, bytes):
            raw_data = data
        else:
            raise ValueError("Unable to send type {}".format(type(data)))

        if stanza and log.getEffectiveLevel() <= logging.VERBOSE:
            log.verbose("[STANZA OUT]:\n%s" %
                        stanza.toXml(pprint=True).decode("utf-8"))

        _send(raw_data)

        if stanza:
            for m in self._mixins:
                hook = partial(m.onSend, self, stanza)
                asyncio.ensure_future(self._runMixin(hook))

    async def sendAndWaitIq(self, child_ns, to=None, child_name="query",
                            type="get", raise_on_error=False, timeout=None,
                            id_prefix=None):
        iq = Iq(to=to, type=type, request=(child_name, child_ns),
                id_prefix=id_prefix)
        resp = await self.sendAndWait(iq, raise_on_error=raise_on_error,
                                           timeout=timeout)
        return resp

    async def sendAndWait(self, stanza, raise_on_error=False, timeout=None):
        if not stanza.id:
            stanza.setId()

        xpath = "/%s[@id='%s']" % (stanza.name, stanza.id)
        self.send(stanza)
        resp = await self.wait([(xpath, None)], timeout=timeout)

        if resp.error is not None and raise_on_error:
            raise resp.error
        else:
            return resp

    async def negotiate(self, timeout=None):
        raise NotImplementedError()

    async def wait(self, xpaths, timeout=None):
        """``xpaths`` is a 2-tuple of the form (xpath, nsmap), or a list of
        the same tuples to wait on a choice of matches. The first matched
        stanza is returned. Passing a ``timeout`` argument will raise a
        asyncio.TimeoutError if not matches are found."""
        global stream_wait_met

        if not isinstance(xpaths, list):
            xpaths = [xpaths]

        if timeout is None and self.default_timeout:
            timeout = self.default_timeout

        log.debug("Stream wait for %s [timeout=%s]" % (xpaths, timeout))
        if _ENFORCE_TIMEOUTS and not timeout:
            raise RuntimeError("Timeout not set error")

        fut = _StreamWaitFuture(xpaths)

        # Run thru queue. Note, once a tasklet has seen a stanza it is skipped
        # by _StreamWaitFuture.matchStanza
        for queued_stanza in self._stanza_queue:
            matched = fut.matchStanza(queued_stanza)
            if matched:
                return queued_stanza.stanza

        self._waiter_futures.append(fut)
        try:
            with timedWait() as timer_stat:
                match = await asyncio.wait_for(fut, timeout)

            if stream_wait_met:
                stream_wait_met.update(timer_stat["total"])
                log.debug("Stream wait - time: {:.3f} "
                          "min/max/avg: {:.6f}/{:.6f}/{:.6f}"
                          .format(stream_wait_met.value,
                                  stream_wait_met.min, stream_wait_met.max,
                                  stream_wait_met.average))

            return match
        except asyncio.TimeoutError as ex:
            raise asyncio.TimeoutError(
                    "Timeout ({}s) while waiting for xpaths: {}"
                    .format(timeout, xpaths)) from ex
        finally:
            self._waiter_futures.remove(fut)

    # asyncio.Protocol implementation
    def connection_made(self, transport, tls=False):
        log.debug("Connection_made: %s", transport)
        self._transport = transport
        self._tls_active = tls
        signalEvent(self._callbacks, "connected", self, tls)

    def starttls_made(self, transport):
        self.connection_made(transport, tls=True)

    async def _handleStanza(self, stanza):

        if isinstance(stanza, stanzas.StreamError):
            signalEvent(self._callbacks, "streamError", self, stanza)
            self._transport.close()
            return

        for m in self._mixins:
            hook = partial(m.onStanza, self, stanza)
            asyncio.ensure_future(self._runMixin(hook))

        self._stanza_queue.append(QueuedStanza(stanza))

        if self._waiter_futures:
            for queued_stanza in self._stanza_queue:
                for fut in [f for f in self._waiter_futures if not f.done()]:
                    matched = fut.matchStanza(queued_stanza)
                    if matched:
                        # XXX: How useful is this since _stanza_queue?
                        # Yield the event loop, which is essential for a handle
                        # and wait in quick succession.
                        await asyncio.sleep(0)

    # asyncio.Protocol implementation
    def data_received(self, data):
        log.debug('[BYTES IN]: {!r}'.format(data.decode()))
        self._parser_task.parse(data)

    # asyncio.Protocol implementation
    def connection_lost(self, reason):
        self._transport = None
        self._tls_active = False
        log.debug('The server closed the connection: %s' % str(reason))
        signalEvent(self._callbacks, "disconnected", self, reason)

    @property
    def default_timeout(self):
        return self._default_timeout

    @default_timeout.setter
    def default_timeout(self, t):
        if t is not None:
            t = int(t)
        self._default_timeout = t

    async def _runMixin(self, functor):
        try:
            await functor()
        except:
            log.exception("{} mixin error".format(functor.__class__.__name__))


class Mixin(object):
    def __init__(self, export_tuples=None):
        """
        ``export_tuples`` is a list of 2-tuples (name, obj) that added to the
        stream object's __dict__, as in __dict__[name] = obj.  By default no
        values are exported.
        """
        self._exports = export_tuples if export_tuples else []

    async def postSession(self, stream):
        """Called after stream negotiation and session creation."""
        pass

    async def onStanza(self, stream, stanza):
        """Called for each incoming Stanza.

        See :func:`vexmpp.utils.xpathFilter` for a decorator that can filter
        only the stanzas the implementation is interested in.
        """
        pass

    async def onSend(self, stream, stanza):
        """Called for each outgoing stanza."""
        pass


class StreamCallbacks:
    def connected(self, stream, tls_active):
        pass

    def disconnected(self, stream, reason):
        pass

    def streamError(self, stream, error):
        pass


class _StreamWaitFuture(asyncio.Future):
    def __init__(self, xpaths, *args, loop=None):
        super().__init__(*args, loop=loop)
        self._xpaths = xpaths
        self._task = asyncio.Task.current_task()

    def matchStanza(self, queued_stanza):
        log.debug(f"MatchStanza: {queued_stanza.stanza.toXml()} xpaths: "
                  "{0} - @{1}".format(self._xpaths, id(self._task)))
        if self._task in queued_stanza.task_set:
            # seen this...
            return False
        queued_stanza.task_set.add(self._task)

        stanza = queued_stanza.stanza
        for xp, nsmap in self._xpaths:
            log.debug("MatchStanza: Testing xpath {} against stanza {}"
                      .format((xp, nsmap), stanza.toXml()))
            if stanza.xml.xpath(xp, namespaces=nsmap):
                log.debug("MatchStanza: matched")
                self.set_result(stanza)
                return True
        log.debug("MatchStanza: NOT matched")
        return False



# -*- coding: utf-8 -*-
import asyncio
from operator import itemgetter
from ..stream import Mixin
from ..stanzas import Presence
from ..utils import xpathFilter
from .muc import NS_URI_USER as MUC_USER_NS

PRES_SUB_XPATH = "/presence[@type='subscribe']"
PRES_SUBED_XPATH = "/presence[@type='subscribed']"
PRES_UNSUB_XPATH = "/presence[@type='unsubscribe']"
PRES_UNSUBED_XPATH = "/presence[@type='unsubscribed']"

S10N_XPATHS = [(PRES_SUB_XPATH, None),
               (PRES_SUBED_XPATH, None),
               (PRES_UNSUB_XPATH, None),
               (PRES_UNSUBED_XPATH, None)]


class SubscriptionAckMixin(Mixin):
    '''
    A stream Mixin which acknowledges subscribe and unsubscribe presence
    request.  By default this class will accept all four subscription types.
    '''

    @xpathFilter(S10N_XPATHS)
    async def onStanza(self, stream, stanza):
        if stanza.xml.xpath(PRES_SUB_XPATH):
            stream.send(Presence(to=stanza.frm,
                                 type=Presence.TYPE_SUBSCRIBED))
        elif stanza.xml.xpath(PRES_UNSUB_XPATH):
            stream.send(Presence(to=stanza.frm,
                                 type=Presence.TYPE_UNSUBSCRIBED))
        elif stanza.xml.xpath(PRES_SUBED_XPATH):
            stream.send(Presence(to=stanza.frm,
                                 type=Presence.TYPE_SUBSCRIBE))
        elif stanza.xml.xpath(PRES_UNSUBED_XPATH):
            stream.send(Presence(to=stanza.frm,
                                 type=Presence.TYPE_UNSUBSCRIBE))


class DenySubscriptionAckMixin(Mixin):
    '''
    A stream Mixin which acknowledges subscribe and unsubscribe presence
    stanzas. By default this class will deny all four subscription types.
    '''

    @xpathFilter(S10N_XPATHS)
    async def onStanza(self, stream, stanza):
        if stanza.xml.xpath(PRES_SUB_XPATH):
            stream.send(Presence(to=stanza.frm,
                                 type=Presence.TYPE_UNSUBSCRIBED))
        elif stanza.xml.xpath(PRES_UNSUB_XPATH):
            stream.send(Presence(to=stanza.frm,
                                 type=Presence.TYPE_SUBSCRIBED))
        elif stanza.xml.xpath(PRES_SUBED_XPATH):
            stream.send(Presence(to=stanza.frm,
                                 type=Presence.TYPE_UNSUBSCRIBE))
        elif stanza.xml.xpath(PRES_UNSUBED_XPATH):
            stream.send(Presence(to=stanza.frm,
                                 type=Presence.TYPE_SUBSCRIBE))


def subscribe(stream, jid, status=None):
    stream.send(Presence(to=jid, type=Presence.TYPE_SUBSCRIBE, status=status))


def unsubscribe(stream, jid):
    stream.send(Presence(to=jid, type=Presence.TYPE_UNSUBSCRIBE))


def authorize(stream, jid):
    stream.send(Presence(to=jid, type=Presence.TYPE_SUBSCRIBED))


def subscribed(stream, jid):
    stream.send(Presence(to=jid, type=Presence.TYPE_SUBSCRIBED))


def deauthorize(stream, jid):
    stream.send(Presence(to=jid, type=Presence.TYPE_UNSUBSCRIBED))


def unsubscribed(stream, jid):
    stream.send(Presence(to=jid, type=Presence.TYPE_UNSUBSCRIBED))


class PresenceDict(dict):
    '''
    A dictionary that must maintain the form:
    {bare-jid: {resource1: presence,
                resource2: presence}}

    That is a contacts bare JID with the value being another dictionary
    containing the last seen presence stanza for each resource.
    '''
    def __init__(self):
        dict.__init__(self)

    def iterpres(self, jid, highest_to_lowest=True):
        jid_key = jid.bare

        # Sorting in the Prensence stanza which uses priority and show.
        jid_pres = sorted([(r, p) for r, p in self[jid_key].items()],
                          key=itemgetter(1))
        for r, p in (reversed(jid_pres) if highest_to_lowest else jid_pres):
            yield r, p


class PresenceCacheMixin(Mixin, PresenceDict):
    '''Mixin and cache (i.e. PresenceDict) for user (non multi-user-chat)
    presence.'''

    def __init__(self):
        PresenceDict.__init__(self)
        Mixin.__init__(self, [('presence_cache', self)])
        self._presence_update_event = DataEvent()
        self._self_presence = None

    async def waitForUpdate(self, timeout=None):
        pres = await asyncio.wait_for(self._presence_update.wait(),
                                      timeout=timeout)
        return pres

    @xpathFilter("/presence")
    async def onStanza(self, stream, presence):

        if (presence.type not in [Presence.TYPE_AVAILABLE,
                                  Presence.TYPE_UNAVAILABLE] or
                presence.xml.xpath("//ns:x", namespaces={"ns": MUC_USER_NS})):
            return

        from_jid = presence.frm
        # Presence from a service, not tracking these
        if not from_jid.resource:
            return

        # Update presence cache
        if from_jid.bare not in self:
            resources = {}
            self[from_jid.bare] = resources
        else:
            resources = self[from_jid.bare]

        updated_stanza = None

        if presence.type != Presence.TYPE_UNAVAILABLE:
            # Add/update resource
            resources[from_jid.resource] = presence
            updated_stanza = presence
        elif from_jid.resource in resources:
            # Unavailable, remove
            del resources[from_jid.resource]
            if len(resources.keys()) == 0:
                del self[from_jid.bare]
            updated_stanza = presence

        if updated_stanza:
            self._presence_update_event.set(presence)
            self._presence_update_event.clear()

        if from_jid == stream.jid:
            self._self_presence = presence

        return presence

    @property
    def self(self):
        return self._self_presence


# FIXME: move me
class DataEvent(asyncio.Event):
    def __init__(self, *, loop=None):
        super().__init__(loop=loop)
        self._data = None

    def set(self, payload):
        self._data = payload
        super().set()

    def clear(self):
        self._data = None
        super().clear()

    async def wait(self):
        await super().wait()
        return self._data



# -*- coding: utf-8 -*-

XML_NS_URI = "http://www.w3.org/XML/1998/namespace"
STREAM_NS_URI = "http://etherx.jabber.org/streams"
STREAM_ERROR_NS_URI = "urn:ietf:params:xml:ns:xmpp-streams"
STANZA_ERROR_NS_URI = "urn:ietf:params:xml:ns:xmpp-stanzas"
CLIENT_NS_URI = "jabber:client"
SERVER_NS_URI = "jabber:server"



# -*- coding: utf-8 -*-
from xml.parsers import expat
from lxml import etree
from collections import deque

from . import getLogger
from .namespaces import STREAM_NS_URI, CLIENT_NS_URI, SERVER_NS_URI


class ParseError(RuntimeError):
    def __init__(self, msg, state):
        super(ParseError, self).__init__(msg)
        self.parser_state = state


class Parser(object):
    def __init__(self):
        self._expat = None
        self._curr = None

        self._stanzas = deque()

        class ParseState(object):
            def __init__(self):
                self.last_data = b""
                self.end_of_stream = False
                self.reset()

            def reset(self):
                self.nsmap = {}
                self.level = 0
                self.elem = None
        self._curr = ParseState()

        self.reset()

    def reset(self):
        log.debug("Resetting expat parser")

        self._expat = expat.ParserCreate("utf-8", " ")
        self._expat.StartNamespaceDeclHandler = self._onStartNamespace
        self._expat.StartElementHandler = self._onStartElement
        self._expat.EndElementHandler = self._onEndElement
        self._expat.CharacterDataHandler = self._onCharData

        self._curr.reset()

    def _onStartNamespace(self, prefix, uri):
        self._curr.nsmap[prefix] = uri

    @staticmethod
    def isStreamHeader(e):
        return e.tag == "{%s}stream" % STREAM_NS_URI

    @staticmethod
    def isStreamError(e):
        return e.tag == "{%s}error" % STREAM_NS_URI

    def parse(self, data):
        if self._curr.end_of_stream:
            # Peer sent </stream:stream>
            self._curr.reset()

        self._curr.last_data += data
        try:
            self._expat.Parse(data)
        except expat.ExpatError as ex:
            raise ParseError(str(ex), self._curr)
        else:
            self._curr.last_data = b""  # No ExpatError, clear state buffer

        parsed_stanzas = []
        if self._stanzas:
            parsed_stanzas += [s for s in self._stanzas]
            self._stanzas.clear()
        return parsed_stanzas

    def _onStartStreamElement(self, name):
        assert(self._curr.elem is None)

        nsmap = {"stream": STREAM_NS_URI}
        if name == "stream":
            nsmap[None] = self._curr.nsmap[None]
        tag = "{%s}%s" % (STREAM_NS_URI, name)

        elem = etree.Element(tag, nsmap=nsmap)

        return elem

    def _onStartStanzaElement(self, name, ns):
        assert(self._curr.elem is None)

        nsmap = None if ns in [CLIENT_NS_URI, SERVER_NS_URI] else {None: ns}
        tag = "{%s}%s" % (ns, name) if nsmap else name

        elem = etree.Element(tag, nsmap=nsmap)

        return elem

    def _onStartChildElement(self, name, ns):
        assert(self._curr.elem is not None)

        nsmap = None if ns in [CLIENT_NS_URI, SERVER_NS_URI] else {None: ns}
        tag = "{%s}%s" % (ns, name) if nsmap else name

        elem = etree.Element(tag, nsmap=nsmap)
        return elem

    def _onStartElement(self, name, attrs):
        # name is 'ns<space>name'
        elem_ns, elem_name = name.split()

        if elem_ns == STREAM_NS_URI:
            elem = self._onStartStreamElement(elem_name)
        elif self._curr.elem is None:
            elem = self._onStartStanzaElement(elem_name, elem_ns)
        else:
            elem = self._onStartChildElement(elem_name, elem_ns)

        for a in attrs:
            if " " in a:
                # XXX: should these even be supported?
                # Namespaced attribute, a is "ns<space>name"
                a_ns, a = a.split()
                elem.attrib["{%s}%s" % (a_ns, a)] = attrs["%s %s" % (a_ns, a)]
            else:
                elem.attrib[a] = attrs[a]

        if self.isStreamHeader(elem):
            self._curr.elem = None
            self._stanzas.append(elem)
        else:
            self._curr.level += 1
            if self._curr.elem is not None:
                self._curr.elem.append(elem)
            self._curr.elem = elem

    def _onEndElement(self, name):
        if tuple(name.split()) == (STREAM_NS_URI, "stream"):
            self._curr.end_of_stream = True
            return

        self._curr.level -= 1
        if self._curr.level == 0:
            self._stanzas.append(self._curr.elem)
            self._curr.reset()
        else:
            if self._curr.elem is not None:
                self._curr.elem = self._curr.elem.getparent()

    def _onCharData(self, data):
        if self._curr.elem is not None and len(data.strip()):
            if self._curr.elem.text is None:
                self._curr.elem.text = data
            else:
                self._curr.elem.text += data


log = getLogger(__name__)



# -*- coding: utf-8 -*-
import uuid
import functools
from copy import deepcopy
from lxml import etree
from .namespaces import (XML_NS_URI, STREAM_NS_URI,
                         CLIENT_NS_URI, SERVER_NS_URI,
                         STANZA_ERROR_NS_URI, STREAM_ERROR_NS_URI)
from .jid import Jid

XML_LANG = "{%s}lang" % XML_NS_URI
STANZA_ERROR_TAG = "{%s}error" % STANZA_ERROR_NS_URI
STREAM_ERROR_TAG = "{%s}error" % STREAM_ERROR_NS_URI


class ElementWrapper:
    _NEXT_ID = 1
    _UUID = str(uuid.uuid4()).split("-")[0]

    def __init__(self, xml):
        if isinstance(xml, ElementWrapper):
            self.xml = xml.xml
        else:
            self.xml = xml

    def _getChildText(self, child):
        e = self.xml.xpath("child::%s" % child)
        return e[0].text if e else None

    def _setChildText(self, child, s):
        matches = self.xml.xpath("child::%s" % child)
        e = matches[0] if matches else None

        if s is None:
            if e is not None:
                self.xml.remove(e)
            return

        if e is None:
            e = etree.Element(child)
            self.xml.append(e)
        e.text = s

    def toXml(self, pprint=False, encoding="utf-8"):
        return etree.tostring(self.xml, pretty_print=pprint, encoding=encoding)

    @property
    def name(self):
        return self.xml.tag

    @property
    def type(self):
        return self.get("type")

    @type.setter
    def type(self, t):
        self.set("type", t)

    @property
    def id(self):
        return self.get("id")

    @id.setter
    def id(self, i):
        self.set("id", i)

    def setId(self, prefix=None):
        id_str = ""

        if prefix and ':' in prefix:
            raise ValueError("Prefix cannot contain ':'")
        elif prefix:
            id_str += "%s:" % prefix

        id_str += "%s-" % ElementWrapper._UUID
        id_str += str(ElementWrapper._NEXT_ID)
        ElementWrapper._NEXT_ID += 1
        self.id = id_str

    def x(self, ns):
        x = self.getChild("x", ns)
        return ElementWrapper(x) if (x is not None) else None

    def getChild(self, name, ns):
        child = self.xml.find("{%s}%s" % (ns, name))
        return child

    # ----- etree Element interface begin -----
    def find(self, *args, **kwargs):
        e = self.xml.find(*args, **kwargs)
        return ElementWrapper(e) if e is not None else None

    def xpath(self, *args, **kwargs):
        matches = self.xml.xpath(*args, **kwargs)
        retval = []
        for m in matches:
            if isinstance(m, str):
                retval.append(m)
            else:
                retval.append(ElementWrapper(m))
        return retval

    def get(self, key, default=None, as_jid=False):
        value = self.xml.get(key, default=default)
        if value:
            return value if not as_jid else Jid(value)
        else:
            return None

    def set(self, attr, s):
        if not s:
            if attr in self.xml.attrib:
                del self.xml.attrib[attr]
        else:
            if isinstance(s, Jid):
                s = s.full
            self.xml.attrib[attr] = s

    def append(self, child_elem):
        if isinstance(child_elem, ElementWrapper):
            child_elem = child_elem.xml
        return self.xml.append(child_elem)

    def remove(self, child_elem):
        if isinstance(child_elem, ElementWrapper):
            child_elem = child_elem.xml
        return self.xml.remove(child_elem)

    def findtext(self, *args, **kwargs):
        return self.xml.findtext(*args, **kwargs)

    def findall(self, *args, **kwargs):
        all_ = self.xml.findall(*args, **kwargs)
        return [ElementWrapper(e) for e in all_]

    def __iter__(self):
        return iter(self.xml)

    @property
    def attrib(self):
        return self.xml.attrib

    @property
    def text(self):
        return self.xml.text

    @text.setter
    def text(self, txt):
        self.xml.text = txt

    @property
    def tag(self):
        return self.xml.tag

    def clear(self):
        return self.xml.clear()

    def getchildren(self):
        return self.xml.getchildren()

    # ----- etree Element interface end -----

    @staticmethod
    def _makeTagName(tag, ns):
        return "{%s}%s" % (ns, tag)

    def appendChild(self, name, ns=None):
        if not ns:
            nsmap = self.xml.nsmap
            # None represents the unprefixed default namespace. Top-level stanza
            # types don't have this.
            ns = nsmap[None] if None in nsmap else None
        else:
            nsmap = {None: ns}

        c = etree.Element("{%s}%s" % (ns, name), nsmap=nsmap)
        self.xml.append(c)
        return ElementWrapper(c)


class Stanza(ElementWrapper):
    XPATH = (None, None)

    TYPE_GET = "get"
    TYPE_SET = "set"
    TYPE_ERROR = "error"
    TYPE_RESULT = "result"

    def __init__(self, tag=None, nsmap=None, xml=None, attrs=None):

        if xml is None and tag:
            xml = etree.Element(tag, nsmap=nsmap)
        elif xml is None:
            raise ValueError("'tag' or 'xml' argument is required")

        super().__init__(xml)

        for name, value in (attrs or {}).items():
            self.set(name, value)

    def _initAttributes(self, to=None, frm=None, type=None, id=None):
        if to:
            self.to = to
        if frm:
            self.frm = frm
        if type:
            self.type = type
        if id:
            self.id = id

    @property
    def to(self):
        return self.get("to", as_jid=True)

    @to.setter
    def to(self, j):
        self.set("to", j)

    @property
    def frm(self):
        return self.get("from", as_jid=True)

    @frm.setter
    def frm(self, j):
        self.set("from", j)

    @property
    def error(self):
        from . import errors
        error = self.xml.xpath("/*/error")
        if error:
            return errors.makeStanzaError(error[0])
        return None

    @error.setter
    def error(self, err):
        from . import errors
        curr = self.xml.xpath("/*/error")
        if curr:
            self.xml.remove(curr[0])

        if err is not None:
            if not isinstance(err, errors.StanzaError):
                raise ValueError("error attribute must be of type StanzaError")
            self.xml.append(err.xml)

    def swapToFrom(self):
        tmp_to = self.to
        tmp_from = self.frm
        if tmp_from:
            self.to = tmp_from
        if tmp_to:
            self.frm = tmp_to

    def errorResponse(self, err):
        err_stanza = deepcopy(self)
        err_stanza.type = "error"
        for c in err_stanza.xml.getchildren():
            err_stanza.xml.remove(c)
        err_stanza.error = err
        err_stanza.swapToFrom()
        return err_stanza

    def resultResponse(self, clear=False):
        res_stanza = deepcopy(self)
        res_stanza.type = "result"
        res_stanza.error = None
        res_stanza.swapToFrom()
        if clear:
            for c in res_stanza.xml.getchildren():
                res_stanza.xml.remove(c)
        return res_stanza


class StreamHeader(Stanza):
    XPATH = ("/stream:stream", {"stream": STREAM_NS_URI})

    def __init__(self, ns=CLIENT_NS_URI, to=None, frm=None, version="1.0",
                 lang="en", id=None, xml=None):
        if xml is not None:
            assert(xml.tag == "{%s}stream" % STREAM_NS_URI)
            assert(xml.nsmap["stream"] == STREAM_NS_URI)
            assert(xml.nsmap[None] in [CLIENT_NS_URI, SERVER_NS_URI])
            super().__init__(xml=xml)
        else:
            assert(ns in [CLIENT_NS_URI, SERVER_NS_URI])
            super().__init__("{%s}stream" % STREAM_NS_URI,
                             nsmap={"stream": STREAM_NS_URI, None: ns})
            self._initAttributes(to=to, frm=frm, id=id)
            self.version = version
            self.lang = lang

    @property
    def version(self):
        return self.get("version")

    @version.setter
    def version(self, v):
        self.set("version", v)

    @property
    def lang(self):
        return self.get(XML_LANG)

    @lang.setter
    def lang(self, l):
        self.set(XML_LANG, l)

    @property
    def defaultNamespace(self):
        return self.xml.nsmap[None]

    def toXml(self, pprint=False, encoding="utf-8"):
        # Special serialization since it must be an open tag.
        header = u"<stream:stream xmlns:stream='%s' xmlns='%s' " % \
                 (STREAM_NS_URI, self.defaultNamespace)
        if self.lang:
            header += u"xml:lang='%s' " % self.lang
        if self.version:
            header += "version='%s'" % self.version
        if self.to:
            header += " to='%s'" % self.to.full
        if self.frm:
            header += " from='%s'" % self.frm.full
        if self.id is not None:
            header += " id='%s'" % self.id
        header += ">\n"
        return header.encode(encoding)


class StreamFeatures(Stanza):
    XPATH = ("/stream:features", {"stream": STREAM_NS_URI})

    def __init__(self, xml=None):
        if xml is not None:
            assert(xml.tag == "{%s}features" % STREAM_NS_URI)
            assert(xml.nsmap["stream"] == STREAM_NS_URI)
            super().__init__(xml=xml)
        else:
            super().__init__("{%s}features" % STREAM_NS_URI,
                             nsmap={"stream": STREAM_NS_URI})

    def getFeature(self, name, ns):
        for feature in self.xml:
            if feature.tag == "{%s}%s" % (ns, name):
                return feature
        return None


class StreamError(Stanza, RuntimeError):
    XPATH = ("/stream:error", {"stream": STREAM_NS_URI})

    def __init__(self, error=None, xml=None):
        assert(error is not None or xml is not None)

        if xml is not None:
            assert(xml.tag == "{%s}error" % STREAM_NS_URI)
            assert(xml.nsmap["stream"] == STREAM_NS_URI)
            super().__init__(xml=xml)
        else:
            super().__init__("{%s}error" % STREAM_NS_URI,
                             nsmap={"stream": STREAM_NS_URI})
            self.error = error

    @property
    def error(self):
        from . import errors
        return errors.makeStreamError(self.xml)

    @error.setter
    def error(self, err):
        from . import errors
        if not isinstance(err, errors.StreamError):
            raise ValueError("error attribute must be of type "
                             "hiss.xmpp.errors.StreamError")
        self.xml = err.xml


class Iq(Stanza):
    XPATH = ("/iq", None)

    def __init__(self, to=None, frm=None, type="get", id=None, request=None,
                 xml=None, id_prefix=None, attrs=None):
        if xml is not None:
            assert(xml.tag == "iq")
            assert(None not in xml.nsmap)
            super().__init__(xml=xml, attrs=attrs)
        else:
            super().__init__("iq", attrs=attrs)
            self._initAttributes(to=to, frm=frm, id=id, type=type)
            if id is None:
                # Iqs most of all need id values
                self.setId(prefix=id_prefix)
            if request:
                name, ns = request
                self.xml.append(etree.Element("{%s}%s" % (ns, name),
                                              nsmap={None: ns}))

    @property
    def request(self):
        for e in self.xml.getchildren():
            if e.tag != STANZA_ERROR_TAG:
                return ElementWrapper(e)
        return None
    query = request


@functools.total_ordering
class Presence(Stanza):
    XPATH = ("/presence", None)

    TYPE_AVAILABLE = 'available'
    TYPE_UNAVAILABLE = 'unavailable'
    TYPE_SUBSCRIBE = 'subscribe'
    TYPE_SUBSCRIBED = 'subscribed'
    TYPE_UNSUBSCRIBE = 'unsubscribe'
    TYPE_UNSUBSCRIBED = 'unsubscribed'
    TYPE_PROBE = 'probe'

    SHOW_AWAY = 'away'
    SHOW_CHAT = 'chat'
    SHOW_DND = 'dnd'
    SHOW_XA = 'xa'

    ORDERED_SHOWS = [SHOW_CHAT, None, SHOW_AWAY, SHOW_XA, SHOW_DND]

    def __init__(self, to=None, frm=None, type=TYPE_AVAILABLE, priority=None,
                 show=None, status=None, xml=None, attrs=None):
        if xml is not None:
            assert(xml.tag == "presence")
            assert(None not in xml.nsmap)
            super().__init__(xml=xml, attrs=attrs)
        else:
            super().__init__("presence", attrs=attrs)
            self._initAttributes(to=to, frm=frm, type=type)
            self.priority = priority
            self.show = show
            self.status = status

    def __gt__(self, rhs):
        # Must implement even with total_ordering to make !lt != gt
        if self < rhs:
            return False
        else:
            if self.priority == rhs.priority and self.show == rhs.show:
                return False
            else:
                return True

    def __lt__(self, rhs):
        if self.priority < rhs.priority:
            return True
        elif self.priority > rhs.priority:
            return False
        else:
            if (self.ORDERED_SHOWS.index(self.show) <=
                    self.ORDERED_SHOWS.index(rhs.show)):
                return False
            else:
                return True

    @property
    def type(self):
        t = self.get("type")
        return t if t else Presence.TYPE_AVAILABLE

    @type.setter
    def type(self, t):
        if t == Presence.TYPE_AVAILABLE:
            if "type" in self.xml.attrib:
                del self.xml.attrib["type"]
        else:
            self.set("type", t)

    @property
    def priority(self):
        t = self._getChildText("priority")
        return int(t) if t is not None else None

    @priority.setter
    def priority(self, i):
        if i is None:
            self._setChildText("priority", None)
        else:
            i = int(i)
            if -128 < i < 127:
                self._setChildText("priority", str(i))
            else:
                raise ValueError("out of range: -128 < priority > 127")

    @property
    def show(self):
        return self._getChildText("show")

    @show.setter
    def show(self, s):
        if s not in self.ORDERED_SHOWS:
            raise ValueError("Invald show: %s" % s)
        self._setChildText("show", s)

    @property
    def status(self):
        return self._getChildText("status")

    @status.setter
    def status(self, s):
        self._setChildText("status", s)


class Message(Stanza):
    XPATH = ("/message", None)

    TYPE_CHAT = "chat"
    TYPE_ERROR = "error"
    TYPE_GC = "groupchat"
    TYPE_HEADLINE = "headline"
    TYPE_NORMAL = "normal"

    def __init__(self, to=None, frm=None, type=TYPE_CHAT, subject=None,
                 body=None, thread=None, xml=None, attrs=None):
        if xml is not None:
            assert(xml.tag == "message")
            assert(None not in xml.nsmap)
            super().__init__(xml=xml, attrs=attrs)
        else:
            super().__init__("message", attrs=attrs)
            self._initAttributes(to=to, frm=frm, type=type)
            self.subject = subject
            self.body = body
            self.thread = thread

    @property
    def type(self):
        t = self.get("type")
        return t if t else Message.TYPE_NORMAL

    @type.setter
    def type(self, t):
        if not t or t == Message.TYPE_NORMAL:
            if "type" in self.xml.attrib:
                del self.xml.attrib["type"]
        else:
            self.set("type", t)

    @property
    def subject(self):
        return self._getChildText("subject")

    @subject.setter
    def subject(self, s):
        self._setChildText("subject", s)

    @property
    def body(self):
        return self._getChildText("body")

    @body.setter
    def body(self, s):
        self._setChildText("body", s)

    @property
    def thread(self):
        return self._getChildText("thread")

    @thread.setter
    def thread(self, s):
        self._setChildText("thread", s)


def makeStanza(elem):
    if elem.tag == "presence":
        return Presence(xml=elem)
    elif elem.tag == "message":
        return Message(xml=elem)
    elif elem.tag == "iq":
        return Iq(xml=elem)
    elif elem.tag == "{%s}stream" % STREAM_NS_URI:
        return StreamHeader(xml=elem)
    elif elem.tag == "{%s}features" % STREAM_NS_URI:
        return StreamFeatures(xml=elem)
    elif elem.tag == "{%s}error" % STREAM_NS_URI:
        return StreamError(xml=elem)
    else:
        return Stanza(xml=elem)



# -*- coding: utf-8 -*-
import asyncio
import functools
from enum import Enum

from .jid import Jid
from .stream import Stream, StreamCallbacks
from .errors import Error
from .utils import signalEvent, resolveHostPort
from .namespaces import CLIENT_NS_URI
from .stanzas import StreamHeader, StreamFeatures, Presence
from .ssl_transport import create_starttls_connection
from .features import sasl, bind, starttls, stream_mgmt
from .protocols import (resourcebind, iqroster, presence, iqversion,
                        entity_time, disco, iqregister, muc)
from . import getLogger

DEFAULT_C2S_PORT = 5222

log = getLogger(__name__)


class TlsOpts(Enum):
    off = 0
    on = 1
    required = 2

    @staticmethod
    def fromString(s):
        for e in TlsOpts:
            if e.name == s:
                return e
        raise ValueError("Invalid TLS option: {}".format(s))


class Credentials(object):
    '''Client credentials, e.g. a user JID and password.'''

    def __init__(self, jid, password):
        if isinstance(jid, str):
            jid = Jid(jid)
        elif not isinstance(jid, Jid):
            raise ValueError("Invalid type: %s" % str(type(jid)))

        self.jid = jid
        self.password = password


class ClientStream(Stream):
    def __init__(self, creds, tls_opt=None, state_callbacks=None,
                 mixins=None, default_timeout=None, register_cb=None):
        self._tls_opt = tls_opt or TlsOpts.on
        self._register_cb = register_cb
        super().__init__(creds, state_callbacks=state_callbacks, mixins=mixins,
                         default_timeout=default_timeout)

    async def _reopenStream(self, timeout=None):
        # Reopen stream
        self._parser_task.reset()
        self.send(StreamHeader(ns=CLIENT_NS_URI, to=self.creds.jid.host))
        # Server sends header <stream:stream>
        await self.wait(StreamHeader.XPATH, timeout)
        # Server stream:features
        features = await self.wait(StreamFeatures.XPATH, timeout)

        return features

    async def negotiate(self, timeout=None):
        # <stream:stream>
        stream_stream = StreamHeader(ns=CLIENT_NS_URI, to=self.creds.jid.host)
        self.send(stream_stream)

        # Server <stream:stream>
        await self.wait(StreamHeader.XPATH, timeout)

        # Server stream:features
        features = await self.wait(StreamFeatures.XPATH, timeout)

        # startttls
        tls_elem = features.getFeature("starttls", starttls.NS_URI)
        if self._tls_opt == TlsOpts.required and tls_elem is None:
            raise Error("TLS required by client but not offered by server.")
        elif (self._tls_opt == TlsOpts.off and tls_elem is not None and
                starttls.isRequired(tls_elem)):
            raise Error("TLS off by client but required by server.")

        if self._tls_opt != TlsOpts.off and tls_elem is not None:
            await starttls.handle(self, tls_elem, timeout=timeout)
            features = await self._reopenStream(timeout=timeout)

        # In-band regisration
        if self._register_cb:
            reg_form = await iqregister.getForm(self)

            self._register_cb(self.creds, reg_form.query)

            reg_form.type = "set"
            reg_form.swapToFrom()
            reg_form.setId()
            await self.sendAndWait(reg_form, timeout=timeout)

        # SASL auth
        mechs_elem = features.getFeature("mechanisms", sasl.NS_URI)
        if mechs_elem is not None:
            await sasl.handle(self, mechs_elem, timeout=timeout)
            features = await self._reopenStream(timeout=timeout)
        else:
            raise Error("Missing mechanisms feature")

        # Resource bind
        bind_elem = features.getFeature("bind", resourcebind.NS_URI)
        if bind_elem is None:
            raise Error("Missing bind feature")
        await bind.handle(self, bind_elem, timeout=timeout)

        # Stream management
        sm_elem = features.getFeature("sm", stream_mgmt.NS_URI)
        if sm_elem is not None and hasattr(self, "stream_mgmt_opts"):
            await stream_mgmt.handle(self, sm_elem,
                                          sm_opts=self.stream_mgmt_opts,
                                          timeout=timeout)
        for mixin in self._mixins:
            await mixin.postSession(self)

    @classmethod
    async def connect(Class, creds, host=None, port=DEFAULT_C2S_PORT,
                loop=None, timeout=None,
                **stream_kwargs):
        '''Connect and negotiate a stream with the server. The connected stream
        is returned.'''
        loop = loop or asyncio.get_event_loop()
        (host,
         port) = await resolveHostPort(host if host else creds.jid.host,
                                            port, loop)
        peer = (host.host, int(port))
        log.verbose("Connecting %s..." % str(peer))

        state_callbacks = stream_kwargs["state_callbacks"] \
                              if "state_callbacks" in stream_kwargs else None
        signalEvent(state_callbacks, "connecting", peer[0], peer[1])

        def _sslContext():
            from OpenSSL.SSL import (Context, SSLv23_METHOD, OP_NO_SSLv2,
                                     OP_NO_SSLv3, VERIFY_PEER)
            ssl_ctx = Context(SSLv23_METHOD)
            ssl_ctx.set_options(OP_NO_SSLv2 | OP_NO_SSLv3)

            def _verifyPeerCb(ctx, x509, errno, errdepth, returncode):
                # FIXME: VERIFY!!!
                print("errno: %s" % errno)
                print("returncode: %s" % returncode)
                print("errdepth: %s" % errdepth)
                print("x509: %s" % x509)
                print("ctx: %s" % ctx)
                return True

            ssl_ctx.set_verify(VERIFY_PEER, _verifyPeerCb)
            return ssl_ctx

        if "mixins" not in stream_kwargs or stream_kwargs["mixins"] is None:
            stream_kwargs["mixins"] = Class.createDefaultMixins()
        ProtocolFactory = functools.partial(Class, creds,
                                            default_timeout=timeout,
                                            **stream_kwargs)

        conn = create_starttls_connection(loop, ProtocolFactory, *peer,
                                          use_starttls=True,
                                          ssl_context=_sslContext(),
                                          server_hostname=creds.jid.host)
        try:
            connected = False
            (transport,
             stream) = await asyncio.wait_for(conn, timeout)

            connected = True
            peer = transport.get_extra_info("peername")

            await stream.negotiate(timeout=timeout)
            signalEvent(state_callbacks, "sessionStarted", stream)
        except Exception as ex:
            if not connected:
                signalEvent(state_callbacks, "connectionFailed",
                            peer[0], peer[1], ex)
            else:
                signalEvent(state_callbacks, "streamError", stream, ex)

            raise

        log.info("%s connected to %s" % (creds.jid.full, peer))
        return stream

    @staticmethod
    def createDefaultMixins():
        return [
                disco.DiscoInfoMixin(),
                iqroster.RosterMixin(),
                iqversion.IqVersionMixin(),
                entity_time.EntityTimeMixin(),
                presence.PresenceCacheMixin(),
                presence.SubscriptionAckMixin(),
                disco.DiscoCacheMixin(),
                muc.MucMixin(),
               ]

    def sendPresence(self, **kwargs):
        self.send(Presence(**kwargs))


class ClientStreamCallbacks(StreamCallbacks):
    def connecting(self, host, port):
        pass

    def connectionFailed(self, host, port):
        pass

    def sessionStarted(self, stream):
        pass



# -*- coding: utf-8 -*-
from ..stanzas import Iq


NS_URI = "urn:ietf:params:xml:ns:xmpp-session"


async def newsession(stream, timeout=None):
    iq = Iq(type="set", request=("session", NS_URI))
    iq.setId("sess")
    resp = await stream.sendAndWait(iq, timeout=timeout)
    if resp.error:
        # TODO
        raise NotImplementedError



# -*- coding: utf-8 -*-
from .. import stream
from ..jid import Jid
from ..stanzas import Iq, ElementWrapper
from ..errors import XmppError
from ..utils import xpathFilter

"""XEP 30"""

from .. import getLogger
log = getLogger(__name__)

NS_URI_BASE = "http://jabber.org/protocol/disco"
NS_URI_INFO = "{}#info".format(NS_URI_BASE)
NS_URI_ITEMS = "{}#items".format(NS_URI_BASE)


async def getInfo(stream, to, node=None, timeout=None):
    iq = Iq(to=to, request=("query", NS_URI_INFO), attrs={"node": node},
            id_prefix="disco#info_get")
    return (await stream.sendAndWait(iq, raise_on_error=True, timeout=timeout))


async def getItems(stream, to, node=None, timeout=None):
    iq = Iq(to=to, request=("query", NS_URI_ITEMS), id_prefix="disco#items_get")
    if node:
        iq.set("node", node)

    iq = await stream.sendAndWait(iq, raise_on_error=True, timeout=timeout)
    return iq


class Identity:
    def __init__(self, category, type, name=None, lang=None):
        self.category = category
        self.type = type
        self.name = name
        self.lang = lang

    def __str__(self):
        return ("Identity [category=%s type=%s name=%s lang=%s]"
                % (self.category, self.type, self.name, self.lang))

    def __hash__(self):
        return self.__str__().__hash__()

    def __eq__(self, o):
        return (type(o) == type(self) and
                o.category == self.category and
                o.type == self.type and
                o.name == self.name and
                o.lang == self.lang)


class Info:
    def __init__(self):
        self.disco_jid = None
        self.identities = set()
        self.features = set()
        self.items = set()
        self.node = None


class DiscoCache:
    def __init__(self):
        self.cache = {}

    def clear(self):
        self.cache.clear()

    def getJidsForFeature(self, feature):
        jids = []
        for disco_jid, disco_info in self.cache.items():
            if feature in disco_info.features:
                jids.append(disco_jid)
        return jids


class DiscoCacheMixin(stream.Mixin):
    def __init__(self):
        self._cache = DiscoCache()
        super().__init__([('disco_cache', self._cache)])

    async def postSession(self, stream):
        await self.update(stream, stream.jid.host)

    async def update(self, stream, disco_jid):
        self._cache.clear()

        # Fetch all disco info for the server
        disco_info = await self._disco(stream, disco_jid, True)

        # Fetch details about all the server's items (but not info about each
        # item)
        if disco_info and disco_info.items is not None:
            for jid in disco_info.items:
                try:
                    await self._disco(stream, jid, False)
                except XmppError as ex:
                    log.warn("Stanza error while disco'ing item '{}': {}"
                             .format(jid.full, ex))

    async def _disco(self, stream, jid, fetch_items):
        if not isinstance(jid, Jid):
            jid = Jid(jid)

        disco_info = Info()
        disco_info.disco_jid = jid

        info = await getInfo(stream, to=jid, timeout=stream.default_timeout)

        for child in info.query:
            if child.tag == "{%s}identity" % NS_URI_INFO:
                ident = Identity(category=child.attrib['category'],
                                 type=child.attrib['type'],
                                 name=child.attrib.get('name', None))
                disco_info.identities.add(ident)
            elif child.tag == "{%s}feature" % NS_URI_INFO:
                disco_info.features.add(child.attrib['var'])

        if fetch_items:
            items = await getItems(stream, jid,
                                   timeout=stream.default_timeout)

            for child in items.query:
                if child.tag == "{%s}item" % NS_URI_ITEMS:
                    disco_info.items.add(Jid(child.attrib['jid']))

        self._cache.cache[jid] = disco_info
        return disco_info


class DiscoInfoMixin(stream.Mixin):
    def __init__(self):
        self._features = []
        super().__init__([('disco_info_features', self._features)])

    @xpathFilter([("/iq[@type='get']/ns:query", {"ns": NS_URI_INFO}),
                  ("/iq[@type='get']/ns:query", {"ns": NS_URI_ITEMS})])
    async def onStanza(self, stream, stanza):
        log.debug("disco#info request")
        if stanza.query.tag.startswith("{%s}" % NS_URI_INFO):
            # disco#info
            query = ElementWrapper(stanza.query)

            ident = query.appendChild("identity")
            ident.set("category", "client")
            ident.set("name", "Botch")
            ident.set("type", "bot")
            query.appendChild("feature").set("var", NS_URI_INFO)
        else:
            # disco#items
            pass

        stream.send(stanza.resultResponse())



"""
:mod:`ssl_transport` --- A transport for asyncio using :mod:`OpenSSL`
#######################################################################

This module provides a socket-based :class:`~asyncio.Transport` for
:mod:`asyncio` which supports deferred TLS as used by the STARTTLS mechanism. In
addition it uses :mod:`OpenSSL` instead of the built-in :mod:`ssl` module, to
provide this and more sophisticated functionality.

The following function can be used to create a connection using the
:class:`STARTTLSTransport`, which itself is documented below:

.. autofunction:: create_starttls_connection

The transport implementation is documented below:

.. autoclass:: STARTTLSTransport(loop, rawsock, protocol, ssl_context, [waiter=None], [use_starttls=False], [post_handshake_callback=None], [peer_hostname=None], [server_hostname=None])
   :members:

"""  # noqa

import asyncio
import socket

from enum import Enum

import OpenSSL.SSL

from . import getLogger
logger = getLogger(__name__)


class _State(Enum):
    RAW_OPEN               = 0x0000  # noqa
    RAW_EOF_RECEIVED       = 0x0001  # noqa

    TLS_HANDSHAKING        = 0x0300  # noqa
    TLS_OPEN               = 0x0100  # noqa
    TLS_EOF_RECEIVED       = 0x0101  # noqa
    TLS_SHUTTING_DOWN      = 0x0102  # noqa
    TLS_SHUT_DOWN          = 0x0103  # noqa

    CLOSED                 = 0x0003  # noqa

    @property
    def eof_received(self):
        return bool(self.value & 0x0001)

    @property
    def tls_started(self):
        return bool(self.value & 0x0100)

    @property
    def tls_handshaking(self):
        return bool(self.value & 0x0200)

    @property
    def is_writable(self):
        return not bool(self.value & 0x0002)

    @property
    def is_open(self):
        return (self.value & 0x3) == 0


class STARTTLSTransport(asyncio.Transport):
    """
    Create a new :class:`asyncio.Transport` which supports TLS and the deferred
    starting of TLS using the :meth:`starttls` method.

    *loop* must be a :class:`asyncio.BaseEventLoop` with support for
    :meth:`BaseEventLoop.add_reader` as well as removal and the writer
    complements.

    *rawsock* must be a :class:`socket.socket` which will be used as the socket
    for the transport. *protocol* must be a :class:`asyncio.Protocol` which will
    be fed the data the transport receives.

    *ssl_context* must be a :class:`OpenSSL.SSL.Context`. It will be used to
    create the :class:`OpenSSL.SSL.Connection` when TLS is enabled on the
    transport.

    *use_starttls* must be a boolean value. If it is true, TLS is not enabled
    immediately. Instead, the user must call :meth:`starttls` to enable TLS on
    the transport. Until that point, the transport is unencrypted. If it is
    false, the TLS handshake is started immediately. This is roughly equivalent
    to calling :meth:`starttls` immediately.

    *peer_hostname* must be either a :class:`str` or :data:`None`. It may be
    used by certificate validators and must be the host name this transport
    actually connected to. That might be (e.g. in the case of XMPP) different
    from the actual domain name the transport communicates with (and for which
    the service must have a valid certificate). This host name may be used by
    certificate validators implementing e.g. DANE.

    *server_hostname* must be either a :class:`str` or :data:`None`. It may be
    used by certificate validators anrd must be the host name for which the peer
    must have a valid certificate (if host name based certificate validation is
    performed). *server_hostname* is also passed via the TLS Server Name
    Indication (SNI) extension if it is given.

    If host names are to be converted to :class:`bytes` by the transport, they
    are encoded using the ``utf-8` codec.

    If *waiter* is not :data:`None`, it must be a :class:`asyncio.Future`. After
    the stream has been established, the futures result is set to a value of
    :data:`None`. If any errors occur, the exception is set on the future.

    If *use_starttls* is true, the future is fulfilled immediately after
    construction, as there is no blocking process which needs to take place. If
    *use_starttls* is false and thus TLS negotiation starts right away, the
    future is fulfilled when TLS negotiation is complete.

    *post_handshake_callback* may be a coroutine or :data:`None`. If it is not
    :data:`None`, it is called asynchronously after the TLS handshake and blocks
    the completion of the TLS handshake until it returns.

    It can be used to perform blocking post-handshake certificate verification,
    e.g. using DANE. The coroutine must not return a value. If it encounters an
    error, an appropriate exception should be raised, which will propagate out
    of :meth:`starttls` and/or passed to the *waiter* future.
    """

    MAX_SIZE = 256 * 1024

    def __init__(self, loop, rawsock, protocol, ssl_context,
                 waiter=None,
                 use_starttls=False,
                 post_handshake_callback=None,
                 peer_hostname=None,
                 server_hostname=None):
        if not use_starttls and not ssl_context:
            raise ValueError("Cannot have STARTTLS disabled (i.e. immediate TLS"
                             " connection) and without SSL context.")

        super().__init__()
        self._rawsock = rawsock
        self._raw_fd = rawsock.fileno()
        self._trace_logger = logger.getChild(
            "trace.fd={}".format(self._raw_fd)
        )
        self._sock = rawsock
        self._protocol = protocol
        self._loop = loop
        self._extra = {
            "socket": rawsock,
        }
        self._waiter = waiter
        self._state = None
        self._conn_lost = 0
        self._buffer = bytearray()
        self._ssl_context = ssl_context
        self._extra.update(
            ssl_context=ssl_context,
            conn=None,
            peername=self._rawsock.getpeername(),
            peer_hostname=peer_hostname,
            server_hostname=server_hostname
        )

        self._paused = False
        self._closing = False

        self._tls_conn = None
        self._tls_read_wants_write = False
        self._tls_write_wants_read = False
        self._tls_post_handshake_callback = post_handshake_callback

        self._state = None
        if not use_starttls:
            self._initiate_tls()
        else:
            self._initiate_raw()

    def _invalid_transition(self, via=None, to=None):
        via_text = (" via {}".format(via)) if via is not None else ""
        to_text = (" to {}".format(to)) if to is not None else ""
        msg = "Invalid state transition (from {}{}{})".format(
            self._state,
            via_text,
            to_text
        )
        logger.error(msg)
        raise RuntimeError(msg)

    def _invalid_state(self, what, exc=RuntimeError):
        msg = "{what} (invalid in state {state}, closing={closing})".format(
            what=what,
            state=self._state,
            closing=self._closing)
        logger.error(msg)
        # raising is optional :)
        return exc(msg)

    def _fatal_error(self, exc, msg):
        if not isinstance(exc, (BrokenPipeError, ConnectionResetError)):
            self._loop.call_exception_handler({
                "message": msg,
                "exception": exc,
                "transport": self,
                "protocol": self._protocol
            })

        self._force_close(exc)

    def _force_close(self, exc):
        self._trace_logger.debug("_force_close called")
        self._remove_rw()
        if self._state == _State.CLOSED:
            # don’t raise here
            raise self._invalid_state("_force_close called")
            return

        self._state = _State.CLOSED

        if self._buffer:
            self._buffer.clear()

        self._loop.remove_reader(self._raw_fd)
        self._loop.remove_writer(self._raw_fd)
        self._loop.call_soon(self._call_connection_lost_and_clean_up, exc)

    def _remove_rw(self):
        self._trace_logger.debug("clearing readers/writers")
        self._loop.remove_reader(self._raw_fd)
        self._loop.remove_writer(self._raw_fd)

    def _call_connection_lost_and_clean_up(self, exc):
        """
        Clean up all resources and call the protocols connection lost method.
        """
        self._state = _State.CLOSED
        try:
            self._protocol.connection_lost(exc)
        finally:
            self._rawsock.close()
            self._tls_conn.set_app_data(None)
            self._tls_conn = None
            self._rawsock = None
            self._protocol = None
            self._loop = None

    def _initiate_raw(self):
        if self._state is not None:
            self._invalid_transition(via="_initiate_raw", to=_State.RAW_OPEN)

        self._state = _State.RAW_OPEN
        self._loop.add_reader(self._raw_fd, self._read_ready)
        self._loop.call_soon(self._protocol.connection_made, self)
        if self._waiter is not None:
            self._loop.call_soon(self._waiter.set_result, None)
            self._waiter = None

    def _initiate_tls(self):
        self._trace_logger.debug("_initiate_tls called")
        if self._state is not None and self._state != _State.RAW_OPEN:
            self._invalid_transition(via="_initiate_tls",
                                     to=_State.TLS_HANDSHAKING)

        self._tls_was_starttls = (self._state == _State.RAW_OPEN)
        self._state = _State.TLS_HANDSHAKING
        self._tls_conn = OpenSSL.SSL.Connection(
            self._ssl_context,
            self._sock)
        self._tls_conn.set_connect_state()
        self._tls_conn.set_app_data(self)
        try:
            self._tls_conn.set_tlsext_host_name(
                self._extra["server_hostname"].encode("IDNA"))
        except KeyError:
            pass
        self._sock = self._tls_conn
        self._extra.update(
            conn=self._tls_conn
        )

        self._tls_do_handshake()

    def _tls_do_handshake(self):
        self._trace_logger.debug("_tls_do_handshake called")
        if self._state != _State.TLS_HANDSHAKING:
            raise self._invalid_state("_tls_do_handshake called")

        try:
            self._tls_conn.do_handshake()
        except OpenSSL.SSL.WantReadError:
            self._trace_logger.debug("registering reader for _tls_do_handshake")
            self._loop.add_reader(self._raw_fd, self._tls_do_handshake)
            return
        except OpenSSL.SSL.WantWriteError:
            self._trace_logger.debug("registering writer for _tls_do_handshake")
            self._loop.add_writer(self._raw_fd, self._tls_do_handshake)
            return
        except Exception as exc:
            self._remove_rw()
            self._fatal_error(exc, "Fatal error on tls handshake")
            if self._waiter is not None:
                self._waiter.set_exception(exc)
            return
        except BaseException as exc:
            self._remove_rw()
            if self._waiter is not None:
                self._waiter.set_exception(exc)
            raise

        self._remove_rw()

        # handshake complete

        self._trace_logger.debug("handshake complete")
        self._extra.update(
            peercert=self._tls_conn.get_peer_certificate()
        )

        if self._tls_post_handshake_callback:
            self._trace_logger.debug("post handshake scheduled via callback")
            task = asyncio.async(self._tls_post_handshake_callback(self))
            task.add_done_callback(self._tls_post_handshake_done)
            self._tls_post_handshake_callback = None
        else:
            self._tls_post_handshake(None)

    def _tls_post_handshake_done(self, task):
        try:
            task.result()
        except BaseException as err:
            self._tls_post_handshake(err)
        else:
            self._tls_post_handshake(None)

    def _tls_post_handshake(self, exc):
        self._trace_logger.debug("_tls_post_handshake called")
        if exc is not None:
            self._fatal_error(exc, "Fatal error on post-handshake callback")
            if self._waiter is not None:
                self._waiter.set_exception(exc)
            return

        self._tls_read_wants_write = False
        self._tls_write_wants_read = False

        self._state = _State.TLS_OPEN

        self._loop.add_reader(self._raw_fd, self._read_ready)
        if self._tls_was_starttls:
            self._loop.call_soon(self._protocol.starttls_made, self)
        else:
            self._loop.call_soon(self._protocol.connection_made, self)
        if self._waiter is not None:
            self._loop.call_soon(self._waiter.set_result, None)

    def _tls_do_shutdown(self):
        self._trace_logger.debug("_tls_do_shutdown called")
        if self._state != _State.TLS_SHUTTING_DOWN:
            raise self._invalid_state("_tls_do_shutdown called")

        try:
            self._sock.shutdown()
        except OpenSSL.SSL.WantReadError:
            self._trace_logger.debug("registering reader for _tls_shutdown")
            self._loop.add_reader(self._raw_fd, self._tls_shutdown)
            return
        except OpenSSL.SSL.WantWriteError:
            self._trace_logger.debug("registering writer for _tls_shutdown")
            self._loop.add_writer(self._raw_fd, self._tls_shutdown)
            return
        except Exception as exc:
            # force_close will take care of removing rw handlers
            self._fatal_error(exc, "Fatal error on tls shutdown")
            return
        except BaseException as exc:
            self._remove_rw()
            raise

        self._remove_rw()
        self._state = _State.TLS_SHUT_DOWN
        # continue to raw shut down
        self._raw_shutdown()

    def _tls_shutdown(self):
        self._state = _State.TLS_SHUTTING_DOWN
        self._tls_do_shutdown()

    def _raw_shutdown(self):
        self._remove_rw()
        self._rawsock.shutdown(socket.SHUT_RDWR)
        self._force_close(None)

    def _read_ready(self):
        if self._state.tls_started and self._tls_write_wants_read:
            self._tls_write_wants_read = False
            self._write_ready()

            if self._buffer:
                self._trace_logger.debug("_read_ready: add writer for more"
                                         " data")
                self._loop.add_writer(self._raw_fd, self._write_ready)

        if self._state.eof_received:
            # no further reading
            return

        try:
            data = self._sock.recv(self.MAX_SIZE)
        except (BlockingIOError, InterruptedError, OpenSSL.SSL.WantReadError):
            pass
        except OpenSSL.SSL.WantWriteError:
            assert self._state.tls_started
            self._tls_read_wants_write = True
            self._trace_logger.debug("_read_ready: swap reader for writer")
            self._loop.remove_reader(self._raw_fd)
            self._loop.add_writer(self._raw_fd, self._write_ready)
        except Exception as err:
            self._fatal_error(err, "Fatal read error on STARTTLS transport")
            return
        else:
            if data:
                self._protocol.data_received(data)
            else:
                keep_open = False
                try:
                    keep_open = bool(self._protocol.eof_received())
                finally:
                    self._eof_received(keep_open)

    def _write_ready(self):
        if self._tls_read_wants_write:
            self._tls_read_wants_write = False
            self._read_ready()

            if not self._paused and not self._state.eof_received:
                self._trace_logger.debug("_write_ready: add reader for more"
                                         " data")
                self._loop.add_reader(self._raw_fd, self._read_ready)

        if self._buffer:
            try:
                nsent = self._sock.send(bytes(self._buffer))
            except (BlockingIOError, InterruptedError,
                    OpenSSL.SSL.WantWriteError):
                nsent = 0
            except OpenSSL.SSL.WantReadError:
                nsent = 0
                assert self._state.tls_started
                self._tls_write_wants_read = True
                self._trace_logger.debug("_write_ready: swap writer for reader")
                self._loop.remove_writer(self._raw_fd)
                self._loop.add_reader(self._raw_fd, self._read_ready)
            except Exception as err:
                self._fatal_error(err,
                                  "Fatal write error on STARTTLS "
                                  "transport")
                return

            if nsent:
                del self._buffer[:nsent]

        if not self._buffer:
            if not self._tls_read_wants_write:
                self._trace_logger.debug("_write_ready: nothing more to write,"
                                   " removing writer")
                self._loop.remove_writer(self._raw_fd)
            if self._closing:
                if self._state.tls_started:
                    self._tls_shutdown()
                else:
                    self._raw_shutdown()

    def _eof_received(self, keep_open):
        self._trace_logger.debug("_eof_received: removing reader")
        self._loop.remove_reader(self._raw_fd)
        if self._state.tls_started:
            if self._tls_conn.get_shutdown() & OpenSSL.SSL.RECEIVED_SHUTDOWN:
                # proper TLS shutdown going on
                if keep_open:
                    self._state = _State.TLS_EOF_RECEIVED
                else:
                    self._tls_shutdown()
            else:
                if keep_open:
                    self._trace_logger.warning("result of eof_received() "
                                               "ignored as shut down is"
                                               " improper")
                self._fatal_error(ConnectionError("Underlying transport "
                                                  "closed"))
        else:
            if keep_open:
                self._state = _State.RAW_EOF_RECEIVED
            else:
                self._raw_shutdown()

    # public API

    def abort(self):
        """
        Immediately close the stream, without sending remaining buffers or
        performing a proper shutdown.
        """
        if self._state == _State.CLOSED:
            self._invalid_state("abort() called")
            return

        self._force_close()

    def can_write_eof(self):
        """
        Return :data:`False`.

        .. note::

           Writing of EOF (i.e. closing the sending direction of the stream) is
           theoretically possible. However, it was deemed by the author that the
           case is rare enough to neglect it for the sake of implementation
           simplicity.

        """
        return False

    def close(self):
        """
        Close the stream. This performs a proper stream shutdown, except if the
        stream is currently performing a TLS handshake. In that case, calling
        :meth:`close` is equivalent to calling :meth:`abort`.

        Otherwise, the transport waits until all buffers are transmitted.
        """

        if self._state == _State.CLOSED:
            self._invalid_state("close() called")
            return

        if self._state == _State.TLS_HANDSHAKING:
            # hard-close
            self._force_close(None)
        elif self._state == _State.TLS_SHUTTING_DOWN:
            # shut down in progress, nothing to do
            pass
        elif self._buffer:
            # there is data to be send left, first wait for it to transmit ...
            self._closing = True
        elif self._state.tls_started:
            # normal TLS state, nothing left to transmit, shut down
            self._tls_shutdown()
        else:
            # normal non-TLS state, nothing left to transmit, close
            self._raw_shutdown()

    def get_extra_info(self, name, default=None):
        """
        The following extra information is available:

        * ``socket``: the underlying :mod:`socket` object
        * ``ssl_context``: the :class:`OpenSSL.SSL.Context` object to use (this
          may be :data:`None` until :meth:`starttls` has been called)
        * ``conn``: :class:`OpenSSL.SSL.Connection` object (:data:`None` if TLS
          is not enabled (yet))
        * ``peername``: return value of :meth:`socket.Socket.getpeername`
        * ``peer_hostname``: The *peer_hostname* value passed to the constructor.
        * ``server_hostname``: The *server_hostname* value passed to the constructor.

        """  # noqa
        return self._extra.get(name, default)

    async def starttls(self, ssl_context=None,
                 post_handshake_callback=None):
        """
        Start a TLS stream on top of the socket. This is an invalid operation if
        the stream is not in RAW_OPEN state.

        If *ssl_context* is set, it overrides the *ssl_context* passed to the
        constructor. If *post_handshake_callback* is set, it overrides the
        *post_handshake_callback* passed to the constructor.
        """
        if self._state != _State.RAW_OPEN or self._closing:
            raise self._invalid_state("starttls() called")

        if ssl_context is not None:
            self._ssl_context = ssl_context
            self._extra.update(
                ssl_context=ssl_context
            )
        if post_handshake_callback is not None:
            self._tls_post_handshake_callback = post_handshake_callback

        self._waiter = asyncio.Future()
        self._initiate_tls()
        try:
            await self._waiter
        finally:
            self._waiter = None

    def write(self, data):
        """
        Write data to the transport. This is an invalid operation if the stream
        is not writable, that is, if it is closed. During TLS negotiation, the
        data is buffered.
        """
        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise TypeError('data argument must be byte-ish (%r)',
                            type(data))

        if not self._state.is_writable or self._closing:
            raise self._invalid_state("write() called")

        if not data:
            return

        if not self._buffer:
            self._loop.add_writer(self._raw_fd, self._write_ready)

        self._buffer.extend(data)

    def write_eof(self):
        """
        Writing the EOF has not been implemented, for the sake of simplicity.
        """
        raise NotImplementedError("Cannot write_eof() on STARTTLS transport")


async def create_starttls_connection(
        loop,
        protocol_factory,
        host=None,
        port=None,
        *,
        sock=None,
        ssl_context=None,
        use_starttls=False,
        **kwargs):
    """
    This is roughly a copy of the asyncio implementation of
    :meth:`asyncio.BaseEventLoop.create_connection`. It returns a pair
    ``(transport, protocol)``, where *transport* is a newly created
    :class:`STARTTLSTransport` instance. The keyword arguments are forwarded to
    the constructor of :class:`STARTTLSTransport`.

    *loop* must be a :class:`asyncio.BaseEventLoop`, with support for
    :meth:`asyncio.BaseEventLoop.add_reader` and the corresponding writer and
    removal functions for sockets.

    *protocol_factory* must be a callable which (without any arguments) returns
    a :class:`asyncio.Protocol` which will be connected to the STARTTLS
    transport.

    *host* and *port* must be a hostname and a port number, or both
    :data:`None`. Both must be :data:`None`, if and only if *sock* is not
    :data:`None`. In that case, *sock* is used instead of a newly created
    socket. *sock* is put into non-blocking mode and must be a stream socket.

    This coroutine returns when the stream is established. If *use_starttls* is
    :data:`False`, this means that the full TLS handshake has to be finished for
    this coroutine to return. Otherwise, no TLS handshake takes place. It must
    be invoked using the :meth:`STARTTLSTransport.starttls` coroutine.
    """

    if host is not None and port is not None:
        host_addrs = await loop.getaddrinfo(
            host, port,
            type=socket.SOCK_STREAM)

        exceptions = []

        for family, type, proto, cname, address in host_addrs:
            sock = None
            try:
                sock = socket.socket(family=family, type=type, proto=proto)
                sock.setblocking(False)
                await loop.sock_connect(sock, address)
            except OSError as exc:
                if sock is not None:
                    sock.close()
                exceptions.append(exc)
            else:
                break
        else:
            if len(exceptions) == 1:
                raise exceptions[0]

            model = str(exceptions[0])
            if all(str(exc) == model for exc in exceptions):
                raise exceptions[0]

            exc = OSError("Multiple exceptions: {}".format(
                ", ".join(map(str, exceptions))))
            exc.exceptions = exceptions
            raise exc
    elif sock is None:
        raise ValueError("sock must not be None if host and/or port are None")
    else:
        sock.setblocking(False)

    protocol = protocol_factory()
    waiter = asyncio.Future(loop=loop)
    transport = STARTTLSTransport(loop, sock, protocol,
                                  ssl_context=ssl_context,
                                  waiter=waiter,
                                  use_starttls=use_starttls,
                                  **kwargs)
    await waiter

    return transport, protocol



# -*- coding: utf-8 -*-
import platform
from lxml import etree

from ..__about__ import __project_name__, __version__
from ..stream import Mixin
from ..utils import xpathFilter

from .. import getLogger
log = getLogger(__name__)


NS_URI = "jabber:iq:version"
GET_XPATH = ("/iq[@type='get']/ns:query", {"ns": NS_URI})


async def get(stream, to, timeout=None):
    iq = await stream.sendAndWaitIq(NS_URI, to=to, raise_on_error=True,
                                    id_prefix="version_get", timeout=timeout)
    return iq


class IqVersionMixin(Mixin):
    '''A stream mixin for handling version IQ 'get' requests.'''

    def __init__(self, name=__project_name__, version=__version__,
                 show_platform=True):
        super().__init__()

        self._name = str(name)
        self._version = str(version)
        self._show_platform = show_platform

    @xpathFilter(GET_XPATH)
    async def onStanza(self, stream, stanza):
        log.debug("jabber:iq:version received")

        result = stanza.resultResponse()

        elem = etree.Element("name")
        elem.text = self._name
        result.query.append(elem)

        elem = etree.Element("version")
        elem.text = self._version
        result.query.append(elem)

        if self._show_platform:
            elem = etree.Element("os")
            elem.text = "%s, %s, %s" % (platform.system(), platform.release(),
                                        platform.version())
            result.query.append(elem)

        stream.send(result)



# -*- coding: utf-8 -*-
import os
import sys
import asyncio
import logging
from getpass import getpass

from nicfit.aio import Application
from nicfit.console import perr

import aiodns
import OpenSSL.SSL

from ..jid import Jid
from .. import version
from ..errors import XmppError
from ..stanzas import Presence
from ..protocols import iqversion, stream_mgmt
from ..client import (Credentials, DEFAULT_C2S_PORT, ClientStream,
                      ClientStreamCallbacks, TlsOpts)

from .util import _outputXml, _register, RegistrationError


async def main(args):
    if "VEX_PASSWD" in os.environ:
        password = os.environ["VEX_PASSWD"]
    else:
        password = getpass("Password for '%s': " % args.jid)
    jid = args.jid
    if not jid.resource:
        jid = Jid((jid.user, jid.host, "vex"))

    tls_opt = TlsOpts.fromString(args.tls)

    mixins = ClientStream.createDefaultMixins()
    if args.stream_mgmt:
        sm_opts = stream_mgmt.Opts()
        mixins.insert(0, stream_mgmt.Mixin(sm_opts))

    try:
        reg_cb = _register if args.register else None
        stream = await ClientStream.connect(Credentials(jid, password),
                                            host=args.host, port=args.port,
                                            state_callbacks=Callbacks(),
                                            tls_opt=tls_opt, mixins=mixins,
                                            register_cb=reg_cb, timeout=5)
    except asyncio.TimeoutError:
        print("Connection timed out", file=sys.stderr)
        return 1
    except XmppError as ex:
        print("XMPP error: %s" % str(ex), file=sys.stderr)
        return 2
    except OpenSSL.SSL.Error as ex:
        print("SSL error: %s" % str(ex), file=sys.stderr)
        return 3
    except aiodns.error.DNSError as ex:
        print("DNS resolution failure.", file=sys.stderr)
        return 5
    except RegistrationError as ex:
        print("Registration error:\n{}".format(ex), file=sys.stderr)
        return 6
    except ConnectionRefusedError as ex:
        print("Connection refused:\n{}".format(ex), file=sys.stderr)
        return 7

    # Server version
    server_version = await iqversion.get(stream, jid.host, timeout=10)
    _outputXml(server_version)

    # Initial presence
    stream.send(Presence())

    if args.disconnect:
        stream.close()
        return 0

    while True:
        try:
            stanza = await stream.wait(("/*", None), timeout=10)
            _outputXml(stanza)
        except asyncio.TimeoutError:
            pass

    return 0


class Callbacks(ClientStreamCallbacks):
    def disconnected(self, stream, reason):
        self._shutdown("Disconnected: {}".format(reason))

    def connectionFailed(self, host, port, reason):
        self._shutdown("Connection failed: {}".format(reason))

    def _shutdown(self, msg):
        global app
        perr(msg)
        app.event_loop.stop()

    def connecting(self, host, port):
        pass

    def sessionStarted(self, stream):
        pass

    def connected(self, stream, tls_active):
        pass

    def streamError(self, stream, error):
        self._shutdown("Stream error: {}".format(error))


about = "Simple XMPP client. The user will be prompted for a login password " \
        "unless the environment variable VEX_PASSWD is set."
app = Application(main, name="vex", description=about, version=version)
app.arg_parser.add_argument("jid", type=Jid, help="Jabber ID for login")
app.arg_parser.add_argument("--register", action="store_true",
                            help="Register for an account before logging in.")
app.arg_parser.add_argument("--host", help="Alternative server for connecting")
app.arg_parser.add_argument("--port", type=int, default=DEFAULT_C2S_PORT,
                            help="Alternative port for connecting")
app.arg_parser.add_argument("--disconnect", action="store_true",
                            help="Disconnect once stream negotiation "
                                 "completes.")
optgroup = app.arg_parser.add_argument_group("Stream feature options")
optgroup.add_argument("--tls", action="store",
                      default="on", choices=[e.name for e in TlsOpts],
                      help="TLS setting, 'on' by default.")
optgroup.add_argument("--stream-mgmt", action="store_true", default=False,
                      help="Enable stream management (XEP 198).")

logging.basicConfig()

if __name__ == "__main__":
    app.run()



# -*- coding: utf-8 -*-
import vexmpp
"""
test_vexmpp
----------------------------------

Tests for `vexmpp` module.
"""


def test_metadata():
    assert vexmpp.version
    assert vexmpp.__about__.__license__
    assert vexmpp.__about__.__project_name__
    assert vexmpp.__about__.__author__
    assert vexmpp.__about__.__author_email__
    assert vexmpp.__about__.__version__
    assert vexmpp.__about__.__version_info__
    assert vexmpp.__about__.__release__
    assert vexmpp.__about__.__version_txt__



# -*- coding: utf-8 -*-
from lxml.etree import Element
from ..stanzas import XML_LANG
from ..utils import stripNsFromTag
from ..errors import XmppError

NS_URI = "urn:ietf:params:xml:ns:xmpp-sasl"


class SaslError(XmppError):
    '''
    Conditions: aborted, account-disabled, credentials-expired,
                encryption-required, incorrect-encoding, invalid-authzid,
                invalid-mechanism, mechanism-too-weak, not-authorized,
                temporary-auth-failure
    '''
    def __init__(self, cond=None, text=None, lang=None, xml=None):
        self.cond = cond
        self.text = text
        self.lang = lang

        if xml is not None:
            # 'xml' overrided and keyword args
            if xml.tag != "{%s}failure" % NS_URI:
                raise ValueError("'xml' is not a SASL error element")
            for child in list(xml):
                if child.tag == "text":
                    self.text = child.text
                    try:
                        self.lang = child.attrib[XML_LANG]
                    except KeyError:
                        pass
                else:
                    self.cond = stripNsFromTag(child.tag, NS_URI)

        if not self.cond:
            raise ValueError("Error condition 'cond' required")

    @property
    def xml(self):
        nsmap = {None: NS_URI}

        e = Element("{%s}failure" % NS_URI, nsmap=nsmap)
        e.append(Element("{%s}%s" % (NS_URI, self.cond), nsmap=nsmap))
        if self.text:
            txt = Element("{%s}text" % NS_URI, nsmap=nsmap)
            if self.lang:
                txt.attrib[XML_LANG] = self.lang
            txt.text = self.text
            e.append(txt)

        return e



# -*- coding: utf-8 -*-
from ..stanzas import ElementWrapper

NS_URI = "jabber:x:data"


class XdataForm(ElementWrapper):
    '''XEP-0004'''

    def __init__(self, xml=None):
        if xml is not None:
            if xml.tag != "{%s}x" % NS_URI:
                raise ValueError("xml is not x-data")
        else:
            from lxml import etree
            xml = etree.Element("{%s}x" % NS_URI, nsmap={None: NS_URI})
            xml.set("type", "form")

        super().__init__(xml)

    @property
    def title(self):
        self._getChildText("title")

    @title.setter
    def title(self, t):
        self._setChildText("title", t)

    @property
    def instructions(self):
        self._getChildText("instructions")

    @instructions.setter
    def instructions(self, t):
        self._setChildText("instructions", t)

    def field(self, var):
        return self.find("./{field}[@var='{var}']"
                         .format(var=var, field=self._makeTagName("field",
                                                                  NS_URI)))

    def appendListField(self, var, options, default=None, multi=False):
        assert(var)
        if default and default not in options:
            raise ValueError("Default value '{}' not in options."
                             .format(default))

        f = self.appendChild("field")
        f.attrib["var"] = var
        f.attrib["type"] = "list-single" if not multi else "list-multi"
        if default:
            f.appendChild("value").text = default
        for val in options:
            f.appendChild("option").appendChild("value").text = val

        return f

    def setValue(self, field, val):
        self.field(field)._setChildText("value", val)



# -*- coding: utf-8 -*-
import time
import random
import asyncio
import functools
from operator import attrgetter
from ipaddress import ip_address

import aiodns

from nicfit import getLogger
log = getLogger(__name__)


class benchmark(object):
    '''A context manager for taking timing blocks of code.'''

    def __init__(self, name=None):
        '''If *name* is provided the ``__exit__`` method will print this string
        along with the total elapsed time.'''
        self.name = name
        self.timer_stats = {}

    def __enter__(self):
        '''Returns a dict with the "start" time. By the __exit__ returns the
        keys "end" and "total" contain those times.'''
        self.start = time.time()
        self.timer_stats["start"] = self.start
        return self.timer_stats

    def __exit__(self, ty, val, tb):
        end = time.time()
        self.timer_stats["end"] = end
        self.timer_stats["total"] = end - self.start
        if self.name:
            print("%s : %0.3f seconds" % (self.name, self.timer_stats["total"]))
        return False


def signalEvent(callbacks, event, *args, **kwargs):
    log.debug("Invoking signal %s" % event)
    if callbacks:
        func = getattr(callbacks, event)
        asyncio.get_event_loop().call_soon(
                functools.partial(func, *args, **kwargs))
    else:
        log.debug("No callbacks set")


def stripNsFromTag(tag, ns):
    '''Removes ``ns`` from the lxml fully qualified tag name.
    e.g. {namespace}tagname would return tagname.'''
    ns_prefix = "{%s}" % ns
    if tag.startswith(ns_prefix):
        return tag[len(ns_prefix):]
    else:
        raise ValueError("tag is not in namsepace '%s'" % ns)


def formatTime(seconds, total=None, short=False):
    '''
    Format ``seconds`` (number of seconds) as a string representation.
    When ``short`` is False (the default) the format is:

        HH:MM:SS.

    Otherwise, the format is exacly 6 characters long and of the form:

        1w 3d
        2d 4h
        1h 5m
        1m 4s
        15s

    If ``total`` is not None it will also be formatted and
    appended to the result seperated by ' / '.
    '''
    def time_tuple(ts):
        if ts is None or ts < 0:
            ts = 0
        hours = ts / 3600
        mins = (ts % 3600) / 60
        secs = (ts % 3600) % 60
        tstr = '%02d:%02d' % (mins, secs)
        if int(hours):
            tstr = '%02d:%s' % (hours, tstr)
        return (int(hours), int(mins), int(secs), tstr)

    if not short:
        hours, mins, secs, curr_str = time_tuple(seconds)
        retval = curr_str
        if total:
            hours, mins, secs, total_str = time_tuple(total)
            retval += ' / %s' % total_str
        return retval
    else:
        units = [
            (u'y', 60 * 60 * 24 * 7 * 52),
            (u'w', 60 * 60 * 24 * 7),
            (u'd', 60 * 60 * 24),
            (u'h', 60 * 60),
            (u'm', 60),
            (u's', 1),
        ]

        seconds = int(seconds)

        if seconds < 60:
            return u'   {0:02d}s'.format(seconds)
        for i in range(len(units) - 1):
            unit1, limit1 = units[i]
            unit2, limit2 = units[i + 1]
            if seconds >= limit1:
                return u'{0:02d}{1}{2:02d}{3}'.format(
                    seconds // limit1, unit1,
                    (seconds % limit1) // limit2, unit2)
        return u'  ~inf'


def xpathFilter(xpaths):
    '''
    FIXME
    '''
    if isinstance(xpaths, str):
        # "xpath" -> [("xpath", None)]
        xpaths = [(xpaths, None)]
    elif isinstance(xpaths, tuple) and isinstance(xpaths[0], str):
        # ("xpath", nsmap) -> [("xpath", nsmap)]
        xpaths = [xpaths]

    def wrapper(func):

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            from .stanzas import Stanza

            stanza = None
            for a in args:
                if isinstance(a, Stanza):
                    stanza = a
                    break
            if stanza is None:
                raise TypeError("No arguments of type Stanza found")
            for xp in xpaths:
                if isinstance(xp, str):
                    xp, ns_map = xp, None
                else:
                    xp, ns_map = xp

                if stanza.xml.xpath(xp, namespaces=ns_map):
                    # Matching xpath, invoke the function
                    return func(*args, **kwargs)

            # No xpaths match, so the decorated function should not be called.
            async def _noOpCoro(*args, **kwargs):
                return None
            return _noOpCoro(*args, **kwargs)

        return wrapped_func

    return wrapper


_dns_cache = {}


async def resolveHostPort(hostname, port, loop, use_cache=True, client_srv=True,
                          srv_records=None, srv_lookup=True):
    global _dns_cache

    def _chooseSrv(_srvs):
        # TODO: random choices based on prio/weight
        return random.choice(_srvs)

    if use_cache and hostname in _dns_cache:
        cached = _dns_cache[hostname]
        if type(cached) is list:
            # Rechoose a SRV record
            srv_choice = _chooseSrv(cached)
            resolved_srv = await resolveHostPort(srv_choice.host,
                                                      srv_choice.port, loop,
                                                      use_cache=True,
                                                      client_srv=client_srv,
                                                      srv_lookup=False)
            return resolved_srv
        else:
            return cached

    resolver = aiodns.DNSResolver(loop=loop)

    ip = None
    try:
        # Given an IP, nothing more to do.
        ip = ip_address(hostname)
        return (str(ip), port)
    except ValueError:
        # Not an IP, move on..,
        pass

    srv_query_type = "_xmpp-client" if client_srv else "_xmpp-server"
    try:
        srv = await resolver.query("{}._tcp.{}".format(srv_query_type,
                                                            hostname),
                                        'SRV')
        srv_results = sorted(srv, key=attrgetter("priority", "weight"))

        if srv_records is not None:
            # Copy to callers list
            srv_records += srv_results

        _dns_cache[hostname] = srv_results
        srv_choice = _chooseSrv(srv_results)

        # Reduce to an IP
        if srv_choice.host != hostname:
            resolved_srv = await resolveHostPort(srv_choice.host,
                                                      srv_choice.port, loop,
                                                      use_cache=use_cache,
                                                      client_srv=client_srv,
                                                      srv_lookup=False)
            _dns_cache[hostname] = resolved_srv
            return resolved_srv

    except aiodns.error.DNSError:
        # No SRV, moving on...
        pass

    # A record
    arecord = await resolver.query(hostname, 'A')

    ip = arecord.pop()
    _dns_cache[hostname] = ip, port
    return ip, port



# -*- coding: utf-8 -*-
from .. import stream
from ..utils import xpathFilter
from ..jid import Jid as BaseJid
from ..stanzas import Presence, Iq, ElementWrapper

'''
Multi-user chat (aka XEP 45)
http://xmpp.org/extensions/xep-0045.html
'''

from .. import getLogger
log = getLogger(__name__)

NS_URI = "http://jabber.org/protocol/muc"
JINC_NS_URI = "http://www.jabber.com/protocol/muc"
NS_URI_ADMIN = "%s#admin" % NS_URI
NS_URI_OWNER = "%s#owner" % NS_URI
NS_URI_UNIQUE = "%s#unique" % NS_URI
NS_URI_USER = "%s#user" % NS_URI
JINC_NS_URI_HISTORY = "%s#history" % JINC_NS_URI
NS_CONFERENCE_URI = "jabber:x:conference"


class MucJid(BaseJid):
    @property
    def nick(self):
        return self.resource

    @property
    def room(self):
        return self.user

    @property
    def room_jid(self):
        return MucJid(self.bare_jid)


def selfPresenceXpath(nick_jid):
    '''110, presence for nick_jid.'''
    xpath = "/presence[@from='{}']/mu:x/mu:status[@code='110']"\
            .format(nick_jid.full)
    return (xpath, {"mu": NS_URI_USER})


def errorPresenceXpath(nick_jid):
    '''Error from the room or service.'''
    xpath = "/presence[@from='{}'][@type='error']".format(nick_jid.bare)
    return (xpath, None)


class RosterItem:
    nickname = None
    affiliation = None
    role = None
    jid = None
    self_presence = False

    def __init__(self, nickname, item_elem=None, self_presence=False):
        self.nickname = nickname
        self.self_presence = self_presence

        def _itemGet(item, attr):
            val = item.get(attr)
            if val in (None, "none"):
                val = None
            return val

        if item_elem is not None:
            self.affiliation = _itemGet(item_elem, "affiliation")
            self.role = _itemGet(item_elem, "role")
            jid = _itemGet(item_elem, "jid")
            self.jid = BaseJid(jid) if jid else None

    def __str__(self):
        return ("RosterItem [nickname: {nickname}, jid: {jid}, "
                "affiliation: {affiliation}, role: {role}]"
                .format(nickname=self.nickname,
                        jid=self.jid.full if self.jid else None,
                        affiliation=self.affiliation, role=self.role))


class RoomInfo:
    '''A container for MUC room information including occupant roster.'''

    def __init__(self, room_jid):
        self.jid = room_jid.room_jid  # Making sure to drop any resource/nick
        self._roster = {}

    def addToRoster(self, roster_item):
        '''Add/update an occupant roster entry corresponding to roster_item.'''
        self.removeFromRoster(roster_item.nickname)
        self._roster[roster_item.nickname] = roster_item

    def removeFromRoster(self, nickname):
        '''Remove an occupant roster entry corresponding to nickname.'''
        item = None
        if nickname in self._roster:
            item = self._roster[nickname]
            del self._roster[nickname]
        return item

    @property
    def self_jid(self):
        for r in self._roster.values():
            if r.self_presence:
                return r.self_presence
        raise AssertionError("No self presence JID")

    @property
    def roster(self):
        '''The roster is a dict with nickname keys and RosterItem values.'''
        return dict(self._roster)


class MucMixin(stream.Mixin):
    '''This class provides MUC room and presence tracking for a stream.'''
    PRESENCE_XPATH = ("/presence/muc_user:x", {"muc_user": NS_URI_USER})

    def __init__(self):
        self._muc_rooms = {}
        super().__init__([('muc_rooms', self._muc_rooms)])

    @xpathFilter(PRESENCE_XPATH)
    async def onStanza(self, stream, stanza):
        log.debug("MucMixin: {}".format(stanza.toXml().decode()))

        pres = stanza
        muc_jid = MucJid(pres.frm)
        room_jid = muc_jid.room_jid
        x = pres.x(NS_URI_USER)
        status = x.find("{%s}status" % NS_URI_USER)
        item = x.find("{%s}item" % NS_URI_USER)

        self_presence = None
        if status is not None and (status.get("code") == "110"):
            self_presence = muc_jid

        if room_jid in self._muc_rooms:
            room_info = self._muc_rooms[room_jid]
        else:
            room_info = RoomInfo(muc_jid)
            self._muc_rooms[room_jid] = room_info

        if pres.type == pres.TYPE_AVAILABLE:
            roster_item = RosterItem(muc_jid.nick, item_elem=item,
                                     self_presence=self_presence)
            room_info.addToRoster(roster_item)
        else:
            room_info.removeFromRoster(muc_jid.nick)
            if self_presence:
                del self._muc_rooms[room_jid]


async def enterRoom(stream, room, service, nick, password=None,
                    config_new_room_callback=None, timeout=None):
    nick_jid = MucJid((room, service, nick))

    room_jid = nick_jid.room_jid

    pres = Presence(to=nick_jid)

    x = pres.appendChild("x", NS_URI)
    if password:
        pw = x.appendChild("password")
        pw.text = password

    stream.send(pres)

    pres = await  stream.wait([selfPresenceXpath(nick_jid),
                              errorPresenceXpath(nick_jid)],
                              timeout=timeout)
    if pres.error:
        raise pres.error

    ROOM_CREATED_PRESENCE_XPATH = ("/presence/mu:x/mu:status[@code='201']",
                                   {"mu": NS_URI_USER})
    ROOM_OWNER_PRESENCE_XPATH = \
            ("/presence/mu:x/mu:item[@affiliation='owner']",
             {"mu": NS_URI_USER})

    # Owner create operations
    if (pres.xml.xpath(ROOM_OWNER_PRESENCE_XPATH[0],
                       namespaces=ROOM_OWNER_PRESENCE_XPATH[1]) and
        pres.xml.xpath(ROOM_CREATED_PRESENCE_XPATH[0],
                       namespaces=ROOM_CREATED_PRESENCE_XPATH[1])):

        if config_new_room_callback is None:
            # New instant room, accept the default config
            iq = Iq(to=room_jid, type="set", request=("query", NS_URI_OWNER))
            x = iq.query.appendChild("x", "jabber:x:data")
            x.attrib["type"] = "submit"
            await stream.sendAndWait(iq, raise_on_error=True, timeout=timeout)
        else:
            # Configure room
            iq = Iq(to=room_jid, type="get", request=("query", NS_URI_OWNER))
            room_config = await stream.sendAndWait(iq, raise_on_error=True,
                                                   timeout=timeout)
            room_config = config_new_room_callback(room_config)
            await stream.sendAndWait(room_config, raise_on_error=True,
                                     timeout=timeout)

    return pres


async def kick(stream, nick_jid, reason, timeout=None):
    iq = Iq(to=nick_jid.room_jid, type="set", request=("query", NS_URI_ADMIN))
    query = ElementWrapper(iq.query)
    item = ElementWrapper(query.appendChild("item"))
    item.set("nick", nick_jid.nick)
    item.set("role", "none")
    query.appendChild("reason").text = reason

    await stream.sendAndWait(iq, raise_on_error=True, timeout=timeout)


async def leaveRoom(stream, nick_jid, leave_msg=None, timeout=None):
    pres = Presence(to=nick_jid, type=Presence.TYPE_UNAVAILABLE)
    if leave_msg:
        pres.appendChild("status").text = leave_msg

    stream.send(pres)
    pres = await stream.wait(selfPresenceXpath(nick_jid))
    return pres



# -*- coding: utf-8 -*-
from lxml import etree
from nicfit.console import pout
try:
    from pygments import highlight
    from pygments.lexers import XmlLexer
    from pygments.formatters import TerminalFormatter
    _has_pygments = True
except:
    _has_pygments = False

from ..protocols import iqregister
from ..protocols.xdata import XdataForm


def _outputXml(stanza):
    xml = stanza.toXml(pprint=True, encoding="utf-8")
    if _has_pygments:
        xml = highlight(xml, XmlLexer(), TerminalFormatter())
    pout(xml)


class RegistrationError(Exception):
    pass


def _register(creds, reg_query):
    oob = reg_query.find("{jabber:x:oob}x")
    if oob is not None:
        inst = reg_query.find("{%s}instructions" % iqregister.NS_URI)
        url = oob.find("{jabber:x:oob}url")
        raise RegistrationError("{}\nurl: {}".format(inst.text, url.text))

    xdata = reg_query.find("{jabber:x:data}x")
    if xdata is not None:
        form = XdataForm(xml=xdata.xml)
        getXData(form, creds)
    else:
        username = reg_query.find("{%s}username" % iqregister.NS_URI)
        username.text = creds.jid.user
        password = reg_query.find("{%s}password" % iqregister.NS_URI)
        password.text = creds.password


def getXData(form, creds):
    def _prompt(txt=None):
        if txt:
            return input("{txt}: ".format(**locals()))
        else:
            return input()

    if form.title:
        pout("{:^80}\n{hr:^80}".format(form.title, hr=("=" * len(form.title))))
    if form.instructions:
        pout("Instructions:\n\t{}\n".format(form.instructions))

    def _appendValue(elem, value_txt):
        value_elem = etree.Element("value")
        value_elem.text = value_txt
        elem.append(value_elem)

    for field in form.findall("{jabber:x:data}field"):
        var = field.get("var")
        value = field.find("{jabber:x:data}value")
        pout("{label} ({var}): {value}"
              .format(label=field.get("label"), var=var,
                      value=value.text if value else ""), end="")
        if not field.find("{jabber:x:data}value"):
            if var in ("username", "password"):
                _appendValue(field, creds.jid.user if var == "username"
                                                   else creds.password)
            else:
                resp = _prompt()
                form.setValue(var, resp)
        pout()



# -*- coding: utf-8 -*-
from lxml import etree
from ..jid import Jid
from ..stanzas import Iq

from .. import getLogger
log = getLogger(__name__)

NS_URI = "urn:ietf:params:xml:ns:xmpp-bind"


async def bind(stream, resource, timeout=None):
    iq = Iq(type="set", request=("bind", NS_URI))
    iq.setId("bind")

    if resource:
        resource_elem = etree.Element("resource")
        resource_elem.text = resource
        iq.request.append(resource_elem)

    resp = await stream.sendAndWait(iq, timeout=timeout)
    if resp.error:
        raise resp.error

    # Get the jid out of the response.
    jid = resp.request.findtext("bind:jid", namespaces={"bind": NS_URI})
    if not jid:
        log.error("Resource bind result did not contain a 'jid'")
        return None

    return Jid(jid)



# -*- coding: utf-8 -*-
from lxml.etree import Element
from .namespaces import STREAM_NS_URI, STREAM_ERROR_NS_URI, STANZA_ERROR_NS_URI
from .stanzas import XML_LANG

'''
Concrete :class:StanzaError types.
    * BadRequestStanzaError
    * ConflictStanzaError
    * FeatureNotImplementedStanzaError
    * ForbiddenStanzaError
    * GoneStanzaError
    * InternalServerErrorStanzaError
    * ItemNotFoundStanzaError
    * JidMalformedStanzaError
    * NotAcceptableStanzaError
    * NotAllowedStanzaError
    * NotAuthorizedStanzaError
    * PolicyViolationStanzaError
    * RecipientUnavailableStanzaError
    * RedirectStanzaError
    * RegistrationRequiredStanzaError
    * RemoteServerNotFoundStanzaError
    * RemoteServerTimeoutStanzaError
    * ResourceConstraintStanzaError
    * ServiceUnavailableStanzaError
    * SubscriptionRequiredStanzaError
    * UndefinedConditionStanzaError
    * UnexpectedRequestStanzaError


Concrete :code:StreamError types.
    * BadFormatStreamError
    * BadNamespacePrefixStreamError
    * ConflictStreamError
    * ConnectionTimeoutStreamError
    * HostGoneStreamError
    * HostUnknownStreamError
    * ImproperAddressingStreamError
    * InternalServerErrorStreamError
    * InvalidFromStreamError
    * InvalidNamespaceStreamError
    * InvalidXmlStreamError
    * NotAuthorizedStreamError
    * NotWellFormedStreamError
    * PolicyViolationStreamError
    * RemoteConnectionFailedStreamError
    * ResetStreamError
    * ResourceConstraintStreamError
    * RestrictedXmlStreamError
    * SeeOtherHostStreamError
    * SystemShutdownStreamError
    * UndefinedConditionStreamError
    * UnsupportedEncodingStreamError
    * UnsupportedFeatureStreamError
    * UnsupportedStanzaTypeStreamError
    * UnsupportedVersionStreamError
'''


class Error(RuntimeError):
    '''A generatl exception type for errors in the vexmpp domain.'''


TYPE_AUTH = "auth"
TYPE_CANCEL = "cancel"
TYPE_CONTINUE = "continue"
TYPE_MODIFY = "modify"
TYPE_WAIT = "wait"


# Tuple format (name, type, default-text, default_lang)
STANZA_ERRORS = (
    ("bad-request", TYPE_MODIFY, "", None),
    ("conflict", TYPE_CANCEL, "", None),
    ("feature-not-implemented", TYPE_CANCEL, "", None),
    ("forbidden", TYPE_AUTH, "", None),
    ("gone", TYPE_CANCEL, "", None),
    ("internal-server-error", TYPE_CANCEL, "", None),
    ("item-not-found", TYPE_CANCEL, "", None),
    ("jid-malformed", TYPE_MODIFY, "", None),
    ("not-acceptable", TYPE_MODIFY, "", None),
    ("not-allowed", TYPE_CANCEL, "", None),
    ("not-authorized", TYPE_AUTH, "", None),
    ("policy-violation", TYPE_MODIFY, "", None),
    ("recipient-unavailable", TYPE_WAIT, "", None),
    ("redirect", TYPE_MODIFY, "", None),
    ("registration-required", TYPE_AUTH, "", None),
    ("remote-server-not-found", TYPE_CANCEL, "", None),
    ("remote-server-timeout", TYPE_WAIT, "", None),
    ("resource-constraint", TYPE_WAIT, "", None),
    ("service-unavailable", TYPE_CANCEL, "", None),
    ("subscription-required", TYPE_AUTH, "", None),
    ("undefined-condition", TYPE_AUTH, "", None),
    ("unexpected-request", TYPE_WAIT, "", None),
)

# Tuple format (name, default-text, default_lang)
STREAM_ERRORS = (
    ("bad-format", "", None),
    ("bad-namespace-prefix", "", None),
    ("conflict", "", None),
    ("connection-timeout", "", None),
    ("host-gone", "", None),
    ("host-unknown", "", None),
    ("improper-addressing", "", None),
    ("internal-server-error", "", None),
    ("invalid-from", "", None),
    ("invalid-namespace", "", None),
    ("invalid-xml", "", None),
    ("not-authorized", "", None),
    ("not-well-formed", "", None),
    ("policy-violation", "", None),
    ("remote-connection-failed", "", None),
    ("reset", "", None),
    ("resource-constraint", "", None),
    ("restricted-xml", "", None),
    ("see-other-host", "", None),
    ("system-shutdown", "", None),
    ("undefined-condition", "", None),
    ("unsupported-encoding", "", None),
    ("unsupported-feature", "", None),
    ("unsupported-stanza-type", "", None),
    ("unsupported-version", "", None),
)


class XmppError(RuntimeError):
    '''An error for XMPP logic errors.'''
    cond = None
    text = None
    lang = None
    app_err = None

    def __str__(self):
        return "%s (text: %s [lang: %s]): %s" % (self.cond, self.text,
                                                 self.lang, self.app_err)


class StanzaError(XmppError):
    type = None

    def __init__(self):
        raise NotImplementedError("Instantiate a concrete error type")

    def __str__(self):
        return "%s (type: %s text: %s [lang: %s]): %s" % (self.cond, self.type,
                                                          self.text, self.lang,
                                                          self.app_err)

    @property
    def xml(self):
        assert(self.cond)

        e = Element("error")

        nsmap = {None: STANZA_ERROR_NS_URI}
        e.append(Element("{%s}%s" % (STANZA_ERROR_NS_URI, self.cond),
                         nsmap=nsmap))
        e[0].attrib["type"] = self.type
        if self.text:
            txt = Element("{%s}text" % STANZA_ERROR_NS_URI, nsmap=nsmap)
            if self.lang:
                txt.attrib[XML_LANG] = self.lang
            txt.text = self.text
            e.append(txt)

        if self.app_err:
            e.append(self.app_err)

        return e


class StreamError(XmppError):
    def __init__(self):
        raise NotImplementedError("Instantiate a concrete error type")

    @property
    def xml(self):
        assert(self.cond)

        e = Element("{%s}error" % STREAM_NS_URI,
                    nsmap={"stream": STREAM_NS_URI})

        nsmap = {None: STREAM_ERROR_NS_URI}
        e.append(Element("{%s}%s" % (STREAM_ERROR_NS_URI, self.cond),
                         nsmap=nsmap))
        if self.text:
            txt = Element("{%s}text" % STREAM_ERROR_NS_URI, nsmap=nsmap)
            if self.lang:
                txt.attrib[XML_LANG] = self.lang
            txt.text = self.text
            e.append(txt)

        if self.app_err:
            e.append(self.app_err)

        return e


def _makeClassName(cond, stanza_err):
    '''Simple util for turning stanza/stream error conditions into class
    names.'''
    class_name = ""
    next_leter_cap = True
    for c in cond:
        if next_leter_cap:
            assert(c != '-')
            c = c.upper()
            next_leter_cap = False
        elif c == '-':
            next_leter_cap = True
            continue
        class_name += c

    if stanza_err:
        return "%sStanzaError" % class_name
    else:
        return "%sStreamError" % class_name


# Make concrete error classes types for all the defined errors.
_global_dict = globals()
for _cond, _type, _txt, _lang in STANZA_ERRORS:
    _name = _makeClassName(_cond, True)

    def stanzaErrCtor(self, text=None, type=None, lang=None, app_err=None):
        # Use constructor values or defer to the Class vars
        self.type = type or self.__class__.type
        self.text = text or self.__class__.text
        self.lang = lang or self.__class__.lang
        self.app_err = (app_err if app_err is not None
                                else self.__class__.app_err)

    _global_dict[_name] = type(_name, (StanzaError,),
                               {"cond": _cond,
                               "type": _type,
                               "text": _txt,
                               "lang": _lang,
                               "app_err": None,
                               "__init__": stanzaErrCtor,
                               })
for _cond, _txt, _lang in STREAM_ERRORS:
    _name = _makeClassName(_cond, False)

    def streamErrCtor(self, text=None, lang=None, app_err=None):
        self.text = text or self.__class__.text
        self.lang = lang or self.__class__.lang
        self.app_err = app_err or self.__class__.app_err

    _global_dict[_name] = type(_name, (StreamError,),
                               {"cond": _cond,
                                "text": _txt,
                                "lang": _lang,
                                "app_err": None,
                                "__init__": streamErrCtor,
                               })
del _global_dict


##
# Make a concrete instance of \c StreamError based on \a xml.
#
# \param xml The &lt;stream:error&gt; Element
# \throws ValueError Thrown if \a xml is not a stream:error or lacks
#                    a condition.
# \returns A concrete instance of \c StreamError or
#          \c UndefinedConditionStreamError if the condition is not
#          recognozed.
def makeStreamError(xml):
    return _makeConcreteError(xml)


def makeStanzaError(xml):
    return _makeConcreteError(xml)


def _makeConcreteError(xml):
    stream_error, stanza_error = False, False
    if xml.xpath("/stream:error", namespaces={"stream": STREAM_NS_URI}):
        stream_error = True
        error_ns = STREAM_ERROR_NS_URI
    elif xml.tag == "error":
        stanza_error = True
        error_ns = STANZA_ERROR_NS_URI
    elif (xml.getchildren() and
            xml.getchildren()[0].tag.startswith("{%s}" % STANZA_ERROR_NS_URI)):
        # Many stream features will wrap a stranza error in a feature-specific
        # parent (xep 198, <failed>, e.g.).
        stanza_error = True
        error_ns = STANZA_ERROR_NS_URI
    else:
        raise ValueError("xml must be a stream:error or stanza error (no ns!)")

    cond, type_, text, lang, app_err = None, None, None, None, None
    for child in list(xml):
        if child.tag == "{%s}text" % error_ns and child.text:
            text = child.text
            lang = child.attrib[XML_LANG] if XML_LANG in child.attrib \
                                          else None
        elif child.tag.startswith("{%s}" % error_ns):
            cond = child
            if stanza_error and "type" in xml.attrib:
                type_ = xml.attrib["type"]
        else:
            app_err = child

    if cond is None and app_err is not None:
        # Not properly namespacing the condition is common
        cond = app_err
        cond.tag = "{%s}%s" % (error_ns, cond.tag)

    if stream_error:
        cond = cond.tag[2 + len(STREAM_ERROR_NS_URI):]
    else:
        cond = cond.tag[2 + len(STANZA_ERROR_NS_URI):]
    Class = _makeClassName(cond, stanza_error)

    try:
        Class = globals()[Class]
    except KeyError:
        if stream_error:
            Class = globals()["UndefinedConditionStreamError"]
        else:
            Class = globals()["UndefinedConditionStanzaError"]

    if stream_error:
        return Class(text=text, lang=lang, app_err=app_err)
    else:
        return Class(type=type_, text=text, lang=lang, app_err=app_err)
